# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

"""
vcs support library base class.
"""

from __future__ import absolute_import, print_function, unicode_literals
import os
import logging


__pychecker__ = 'unusednames=spec,url,version,basepath,untracked'


class VcsError(Exception):
    """To be thrown when an SCM Client faces a situation because of a
    violated assumption"""

    def __init__(self, value):
        super(VcsError, self).__init__(value)
        self.value = value

    def __str__(self):
        return repr(self.value)


class VcsClientBase(object):
    """
    parent class for all vcs clients, provides their public API
    """

    def __init__(self, vcs_type_name, path):
        """
        subclasses may raise VcsError when a dependency is missing
        """
        self._path = path
        if path is None:
            raise VcsError("Cannot initialize VCSclient without path")
        self._vcs_type_name = vcs_type_name
        self.logger = logging.getLogger('vcstools')

    @staticmethod
    def get_environment_metadata():
        """
        For debugging purposes, returns a dict containing information
        about the environment, like the version of the SCM client, or
        version of libraries involved.
        Suggest considering keywords "version", "dependency", "features" first.
        :returns: a dict containing relevant information
        :rtype: dict
        """
        raise NotImplementedError(
            "Base class get_environment_metadata method must be overridden")

    def path_exists(self):
        """
        helper function
        """
        return os.path.exists(self._path)

    def get_path(self):
        """
        returns the path this client was configured for
        """
        return self._path

    def url_matches(self, url, url_or_shortcut):
        """
        client can decide whether the url and the other url are equivalent.
        Checks string equality by default
        :param url_or_shortcut: url or shortcut (e.g. bzr launchpad url)
        :returns: bool if params are equivalent
        """
        if url is None or url_or_shortcut is None:
            return False
        return url.rstrip('/') == url_or_shortcut.rstrip('/')

    def get_url(self):
        """
        :returns: The source control url for the path or None if not set
        :rtype: str
        """
        raise NotImplementedError(
            "Base class get_url method must be overridden for client type %s" %
            self._vcs_type_name)

    def get_version(self, spec=None):
        """
        Find an identifier for a the current or a specified
        revision. Token spec might be a tagname, branchname,
        version-id, SHA-ID, ... depending on the VCS implementation.

        :param spec: token for identifying repository revision
        :type spec: str
        :returns: current revision number of the repository.  Or if
          spec is provided, the respective revision number.
        :rtype: str
        """
        raise NotImplementedError("Base class get_version method must be overridden for client type %s " %
                                  self._vcs_type_name)

    def get_current_version_label(self):
        """
        Find an description for the current local version.
        Token spec might be a branchname,
        version-id, SHA-ID, ... depending on the VCS implementation.

        :returns: short description of local version (e.g. branchname, tagename).
        :rtype: str
        """
        raise NotImplementedError("Base class get_current_version method must be overridden for client type %s " %
                                  self._vcs_type_name)

    def get_remote_version(self, fetch=False):
        """
        Find an identifier for the current revision on remote.
        Token spec might be a tagname,
        version-id, SHA-ID, ... depending on the VCS implementation.

        :param fetch: if False, only local information may be used
        :returns: current revision number of the remote repository.
        :rtype: str
        """
        raise NotImplementedError("Base class get_remote_version method must be overridden for client type %s " %
                                  self._vcs_type_name)

    def get_default_remote_version_label(self):
        """
        Find a label for the default branch on remote, meaning
        the one that would be checked out on a clean checkout.

        :returns: a label or None (if not applicable)
        :rtype: str
        """
        raise NotImplementedError("Base class get_default_remote_version_label" +
                                  "method must be overridden for client type %s " %
                                  self._vcs_type_name)

    def checkout(self, url, version=None, verbose=False, shallow=False, timeout=None):
        """
        Attempts to create a local repository given a remote
        url. Fails if a target path exists, unless it's an empty directory.
        If a version is provided, the local repository
        will be updated to that revision. It is possible that
        after a failed call to checkout, a repository still exists,
        e.g. if an invalid revision was given.
        If shallow is provided, the scm client may checkout less
        than the full repository history to save time / disk space.
        If a timeout is specified, any pending operation will fail after
        the specified amount (in seconds). NOTE: this parameter might or
        might not be honored, depending on VCS client implementation.
        :param url: where to checkout from
        :type url: str
        :param version: token for identifying repository revision
        :type version: str
        :param shallow: hint to checkout less than a full repository
        :type shallow: bool
        :param timeout: maximum allocated time to perform operation
        :type shallow: int
        :returns: True if successful
        """
        raise NotImplementedError("Base class checkout method must be overridden for client type %s " %
                                  self._vcs_type_name)

    def update(self, version=None, verbose=False, timeout=None):
        """
        Sets the local copy of the repository to a version matching
        the version parameter. Fails when there are uncommited changes.
        On failures (also e.g. network failure) grants the
        checked out files are in the same state as before the call.
        If a timeout is specified, any pending operation will fail after
        the specified amount (in seconds)

        :param version: token for identifying repository revision
           desired.  Token might be a tagname, branchname, version-id,
           SHA-ID, ... depending on the VCS implementation.
        :param timeout: maximum allocated time to perform operation
        :type shallow: int
        :returns: True on success, False else
        """
        raise NotImplementedError("Base class update method must be overridden for client type %s " %
                                  self._vcs_type_name)

    @staticmethod
    def static_detect_presence(path):
        """For auto detection"""
        raise NotImplementedError(
            "Base class detect_presence method must be overridden")

    def detect_presence(self):
        """For auto detection"""
        # call static method
        return self.static_detect_presence(self._path)

    def get_vcs_type_name(self):
        """ used when auto detected """
        return self._vcs_type_name

    def get_diff(self, basepath=None):
        """
        :param basepath: diff paths will be relative to this, if any
        :returns: A string showing local differences
        :rtype: str
        """
        raise NotImplementedError(
            "Base class get_diff method must be overridden")

    def get_status(self, basepath=None, untracked=False, **kwargs):
        """
        Calls scm status command. Output must be terminated by newline
        unless empty.

        Semantics of untracked are difficult to generalize.
        In SVN, this would be new files only. In git,
        hg, bzr, this would be changes that have not been added for
        commit.

        Extra keyword arguments are passed along to the underlying vcs code.
        See the specific implementations of get_status() for extra options.

        :param basepath: status path will be relative to this, if any
        :param untracked: whether to also show changes that would not commit
        :returns: A string summarizing locally modified files
        :rtype: str
        """
        raise NotImplementedError("Base class get_status method must be overridden for client type %s " %
                                  self._vcs_type_name)

    def get_affected_files(self, revision):
        """
        Get the files that were affected by a specific revision
        :param revision: SHA or revision number.
        :returns: A list of strings with the files affected by a specific commit
        """
        raise NotImplemented(
            "Base class get_affected_files method must be overriden")

    def get_log(self, relpath=None, limit=None):
        """
        Calls scm log command.

        This returns a list of dictionaries with the following fields:
            - id: the commit SHA or revision number
            - date: the date the commit was made (python datetime)
            - author: the name of the author of the commit, if available
            - email: the e-mail address of the author of the commit
            - message: the commit message, if any

        :param relpath: (optional) restrict logs to events on this
        resource path (folder or file) relative to the root of the
        repository. If None (default), this is the root of the
        repository.
        :param limit: (optional) the maximum number of log entries
        that should be retrieved. If None (default), there is no
        limit.
        """
        raise NotImplementedError(
            "Base class get_log method must be overridden")

    def export_repository(self, version, basepath):
        """
        Calls scm equivalent to `svn export`, removing scm meta
        information and tar gzip'ing the repository at a given version
        to the given basepath.

        :param version: version of the repository to export.  This can
        be a branch, tag, or path (svn).  When specifying the version
        as a path for svn, the path should be relative to the root of
        the svn repository, i.e. 'trunk', or 'tags/1.2.3', or './' for
        the root.
        :param basepath: this is the path to the tar gzip, excluding
        the extension which will be .tar.gz
        :returns: True on success, False otherwise.
        """
        raise NotImplementedError("Base class export_repository method must be overridden for client type %s " %
                                  self._vcs_type_name)

    def get_branches(self, local_only=False):
        """
        Returns a list of all branches in the vcs repository.

        :param local_only: if True it will only list local branches
        :returns: list of branches in the repository, [] if none exist
        """
        raise NotImplementedError("Base class get_branches method must "
                                  "be overridden")



from __future__ import print_function
import os
from pkg_resources import parse_version, get_distribution




def test_pep8_conformance():
    """Test source code for PEP8 conformance"""

    try:
        import pep8
    except:
        print("Skipping pep8 Tests because pep8.py not installed.")
        return

    # Skip test if pep8 is not new enough
    pep8_version = parse_version(get_distribution('pep8').version)
    needed_version = parse_version('1.0')
    if pep8_version < needed_version:
        print("Skipping pep8 Tests because pep8.py is too old")
        return

    pep8style = pep8.StyleGuide(max_line_length=120)
    report = pep8style.options.report
    report.start()
    pep8style.options.exclude.append('git_archive_all.py')
    pep8style.input_dir(os.path.join('..', 'vcstools', 'src'))
    report.stop()
    assert report.total_errors == 0, "Found '{0}' code style errors (and warnings).".format(report.total_errors)



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, print_function, unicode_literals
import errno
import os
import sys
import copy
import shlex
import subprocess
import logging
import netrc
import tempfile
import shutil
import threading
import signal

try:
    # py3k
    from urllib.request import urlopen, HTTPPasswordMgrWithDefaultRealm, \
        HTTPBasicAuthHandler, build_opener
    from urllib.parse import urlparse
    from queue import Queue
except ImportError:
    # py2.7
    from urlparse import urlparse
    from urllib2 import urlopen, HTTPPasswordMgrWithDefaultRealm, \
        HTTPBasicAuthHandler, build_opener
    from Queue import Queue

from vcstools.vcs_base import VcsError


def ensure_dir_notexists(path):
    """
    helper function, removes dir if it exists

    :returns: True if dir does not exist after this function
    :raises: OSError if dir exists and removal failed for non-trivial reasons
    """
    try:
        if os.path.exists(path):
            os.rmdir(path)
        return True
    except OSError as ose:
        # ignore if directory
        if ose.errno not in [errno.ENOENT, errno.ENOTEMPTY, errno.ENOTDIR]:
            return False


def urlopen_netrc(uri, *args, **kwargs):
    '''
    wrapper to urlopen, using netrc on 401 as fallback
    Since this wraps both python2 and python3 urlopen, accepted arguments vary

    :returns: file-like object as urllib.urlopen
    :raises: IOError and urlopen errors
    '''
    try:
        return urlopen(uri, *args, **kwargs)
    except IOError as ioe:
        if hasattr(ioe, 'code') and ioe.code == 401:
            # 401 means authentication required, we try netrc credentials
            result = _netrc_open(uri)
            if result is not None:
                return result
        raise


def urlretrieve_netrc(url, filename=None):
    '''
    writes a temporary file with the contents of url. This works
    similar to urllib2.urlretrieve, but uses netrc as fallback on 401,
    and has no reporthook or data option. Also urllib2.urlretrieve
    malfunctions behind proxy, so we avoid it.

    :param url: What to retrieve
    :param filename: target file (default is basename of url)
    :returns: (filename, response_headers)
    :raises: IOError and urlopen errors
    '''
    fname = None
    fhand = None
    try:
        resp = urlopen_netrc(url)
        if filename:
            fhand = open(filename, 'wb')
            fname = filename
        else:
            # Make a temporary file
            fdesc, fname = tempfile.mkstemp()
            fhand = os.fdopen(fdesc, "wb")
            # Copy the http response to the temporary file.
        shutil.copyfileobj(resp, fhand)
    finally:
        if fhand:
            fhand.close()
    return (fname, resp.headers)


def _netrc_open(uri, filename=None):
    '''
    open uri using netrc credentials.

    :param uri: uri to open
    :param filename: optional, path to non-default netrc config file
    :returns: file-like object from opening a socket to uri, or None
    :raises IOError: if opening .netrc file fails (unless file not found)
    '''
    if not uri:
        return None
    parsed_uri = urlparse(uri)
    machine = parsed_uri.netloc
    if not machine:
        return None
    opener = None
    try:
        info = netrc.netrc(filename).authenticators(machine)
        if info is not None:
            (username, _, password) = info
            if username and password:
                pass_man = HTTPPasswordMgrWithDefaultRealm()
                pass_man.add_password(None, machine, username, password)
                authhandler = HTTPBasicAuthHandler(pass_man)
                opener = build_opener(authhandler)
                return opener.open(uri)
        else:
            # caught below, like other netrc parse errors
            raise netrc.NetrcParseError('No authenticators for "%s"' % machine)
    except IOError as ioe:
        if ioe.errno != 2:
            # if = 2, User probably has no .netrc, this is not an error
            raise
    except netrc.NetrcParseError as neterr:
        logger = logging.getLogger('vcstools')
        logger.warn('WARNING: parsing .netrc: %s' % str(neterr))
    # we could install_opener() here, but prefer to keep
    # default opening clean. Client can do that, though.
    return None


def normalized_rel_path(path, basepath):
    """
    If path is absolute, return relative path to it from
    basepath. If relative, return it normalized.

    :param path: an absolute or relative path
    :param basepath: if path is absolute, shall be made relative to this
    :returns: a normalized relative path
    """
    # gracefully ignore invalid input absolute path + no basepath
    if path is None:
        return basepath
    if os.path.isabs(path) and basepath is not None:
        return os.path.normpath(os.path.relpath(os.path.realpath(path), os.path.realpath(basepath)))
    return os.path.normpath(path)


def sanitized(arg):
    """
    makes sure a composed command to be executed via shell was not injected.

    A composed command would be like "ls %s"%foo.
    In this example, foo could be "; rm -rf *"
    sanitized raises an Error when it detects such an attempt

    :raises VcsError: on injection attempts
    """
    if arg is None or arg.strip() == '':
        return ''
    arg = str(arg.strip('"').strip())
    safe_arg = '"%s"' % arg
    # this also detects some false positives, like bar"";foo
    if '"' in arg:
        if (len(shlex.split(safe_arg, False, False)) != 1):
            raise VcsError("Shell injection attempt detected: >%s< = %s" %
                           (arg, shlex.split(safe_arg, False, False)))
    return safe_arg


def _discard_line(line):
    if line is None:
        return True
    # the most common feedback lines of scms. We don't care about those. We let through anything unusual only.
    discard_prefixes = ["adding ", "added ", "updating ", "requesting ", "pulling from ",
                        "searching for ", "(", "no changes found",
                        "0 files",
                        "A  ", "D  ", "U  ",
                        "At revision", "Path: ", "First,",
                        "Installing", "Using ",
                        "No ", "Tree ",
                        "All ",
                        "+N  ", "-D  ", " M  ", " M* ", "RM"  # bzr
                        ]
    for pre in discard_prefixes:
        if line.startswith(pre):
            return True
    return False


def _read_shell_output(proc, no_filter, verbose, show_stdout, output_queue):
    # when we read output in while loop, it would not be returned
    # in communicate()
    stdout_buf = []
    stderr_buf = []
    if not no_filter:
        if (verbose or show_stdout):
            # this loop runs until proc is done it listen to the pipe, print
            # and stores result in buffer for returning this allows proc to run
            # while we still can filter out output avoiding readline() because
            # it may block forever
            for line in iter(proc.stdout.readline, b''):
                line = line.decode('UTF-8')
                if line is not None and line != '':
                    if verbose or not _discard_line(line):
                        sys.stdout.write(line),
                        stdout_buf.append(line)
                if (not line or proc.returncode is not None):
                    break
        # stderr was swallowed in pipe, in verbose mode print lines
        if verbose:
            for line in iter(proc.stderr.readline, b''):
                line = line.decode('UTF-8')
                if line != '':
                    sys.stdout.write(line),
                    stderr_buf.append(line)
                if not line:
                    break
    output_queue.put(proc.communicate())
    output_queue.put(stdout_buf)
    output_queue.put(stderr_buf)


def run_shell_command(cmd, cwd=None, shell=False, us_env=True,
                      show_stdout=False, verbose=False, timeout=None,
                      no_warn=False, no_filter=False):
    """
    executes a command and hides the stdout output, loggs stderr
    output when command result is not zero. Make sure to sanitize
    arguments in the command.

    :param cmd: A string to execute.
    :param shell: Whether to use os shell.
    :param us_env: changes env var LANG before running command, can influence program output
    :param show_stdout: show some of the output (except for discarded lines in _discard_line()), ignored if no_filter
    :param no_warn: hides warnings
    :param verbose: show all output, overrides no_warn, ignored if no_filter
    :param timeout: time allocated to the subprocess
    :param no_filter: does not wrap stdout, so invoked command prints everything outside our knowledge
    this is DANGEROUS, as vulnerable to shell injection.
    :returns: ( returncode, stdout, stderr); stdout is None if no_filter==True
    :raises: VcsError on OSError
    """
    try:
        env = copy.copy(os.environ)
        if us_env:
            env["LANG"] = "en_US.UTF-8"
        if no_filter:
            # in no_filter mode, we cannot pipe stdin, as this
            # causes some prompts to be hidden (e.g. mercurial over
            # http)
            stdout_target = None
            stderr_target = None
        else:
            stdout_target = subprocess.PIPE
            stderr_target = subprocess.PIPE

        # additional parameters to Popen when using a timeout
        crflags = {}
        if timeout is not None:
            if hasattr(os.sys, 'winver'):
                crflags['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP
            else:
                crflags['preexec_fn'] = os.setsid

        proc = subprocess.Popen(cmd,
                                shell=shell,
                                cwd=cwd,
                                stdout=stdout_target,
                                stderr=stderr_target,
                                env=env,
                                **crflags)

        # using a queue to enable usage in a separate thread
        q = Queue()
        if timeout is None:
            _read_shell_output(proc, no_filter, verbose, show_stdout, q)
        else:
            t = threading.Thread(target=_read_shell_output,
                                 args=[proc, no_filter, verbose, show_stdout, q])
            t.start()
            t.join(timeout)
            if t.isAlive():
                if hasattr(os.sys, 'winver'):
                    os.kill(proc.pid, signal.CTRL_BREAK_EVENT)
                else:
                    os.killpg(proc.pid, signal.SIGTERM)
                t.join()
        (stdout, stderr) = q.get()
        stdout_buf = q.get()
        stderr_buf = q.get()

        if stdout is not None:
            stdout_buf.append(stdout.decode('utf-8'))
        stdout = "\n".join(stdout_buf)
        if stderr is not None:
            stderr_buf.append(stderr.decode('utf-8'))
        stderr = "\n".join(stderr_buf)
        message = None
        if proc.returncode != 0 and stderr is not None and stderr != '':
            logger = logging.getLogger('vcstools')
            message = "Command failed: '%s'" % (cmd)
            if cwd is not None:
                message += "\n run at: '%s'" % (cwd)
            message += "\n errcode: %s:\n%s" % (proc.returncode, stderr)
            if not no_warn:
                logger.warn(message)
        result = stdout
        if result is not None:
            result = result.rstrip()
        return (proc.returncode, result, message)
    except OSError as ose:
        logger = logging.getLogger('vcstools')
        message = "Command failed with OSError. '%s' <%s, %s>:\n%s" % (cmd, shell, cwd, ose)
        logger.error(message)
        raise VcsError(message)



#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, print_function, unicode_literals
import os
import unittest
import tarfile
import tempfile
import shutil
import subprocess
import mock

from vcstools.tar import TarClient
from test.mock_server import start_mock_server

def tarfile_contents():
    '''
    :returns: binary string of ROS package-release-like tarfile to serve
    '''
    tar_directory = tempfile.mkdtemp()
    filename = os.path.join(tar_directory, 'sample.tar.gz')
    pkgname = "sample-1.0"
    pkg_directory = os.path.join(tar_directory, pkgname)
    os.mkdir(pkg_directory)
    packagexml = os.path.join(pkg_directory, "package.xml")
    with open(packagexml, 'w+') as f:
        f.write('<?xml version="1.0"?><package>' + ('<name>sample</name>' * 1000) + '</package>')

    with tarfile.open(filename, "w:gz") as tar_handle:
        tar_handle.addfile(tarfile.TarInfo(os.path.join(pkgname, "package.xml")), packagexml)
    tar_handle.close()

    with open(filename, mode='rb') as file: # b is important -> binary
        result = file.read()
    shutil.rmtree(tar_directory)
    return result



class TarClientTest(unittest.TestCase):
    '''Test against mock http server'''

    @classmethod
    def setUpClass(self):
        baseURL = start_mock_server(tarfile_contents())
        self.remote_url = baseURL + '/downloads/1.0.tar.gz'
        self.package_version = "sample-1.0"

    def setUp(self):
        self.directories = {}

    def tearDown(self):
        for d in self.directories:
            self.assertTrue(os.path.exists(self.directories[d]))
            shutil.rmtree(self.directories[d])
            self.assertFalse(os.path.exists(self.directories[d]))

    def test_get_url_by_reading(self):
        directory = tempfile.mkdtemp()
        self.directories['local'] = directory

        local_path = os.path.join(directory, "local")

        client = TarClient(local_path)
        self.assertTrue(client.checkout(self.remote_url, self.package_version))

        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_url(), self.remote_url)
        #self.assertEqual(client.get_version(), self.package_version)

    def test_get_url_nonexistant(self):
        local_path = "/tmp/dummy"
        client = TarClient(local_path)
        self.assertEqual(client.get_url(), None)

    def test_get_type_name(self):
        local_path = "/tmp/dummy"
        client = TarClient(local_path)
        self.assertEqual(client.get_vcs_type_name(), 'tar')


    @mock.patch('netrc.netrc') # cannot rely on local ~/.netrc file
    def test_checkout_parametrized(self, patched_netrc):
        netrc_mock = mock.Mock()
        netrc_mock.authenticators.return_value = ('user', '' , 'password')
        patched_netrc.return_value = netrc_mock
        for query_params in ['', '?chunked=true', '?auth=true', '?chunked=true&auth=true']:
            self.check_checkout(query_params)

    # parametrized
    def check_checkout(self, query_params):
        # checks out all subdirs
        directory = tempfile.mkdtemp()
        self.directories["checkout_test"] = directory
        local_path = os.path.join(directory, "exploration")
        client = TarClient(local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(self.remote_url + query_params))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), local_path)
        self.assertEqual(client.get_url(), self.remote_url + query_params)
        # make sure the tarball subdirectory was promoted correctly.
        self.assertTrue(os.path.exists(os.path.join(local_path,
                                                    self.package_version,
                                                    'package.xml')))

    def test_checkout_dir_exists(self):
        directory = tempfile.mkdtemp()
        self.directories["checkout_test"] = directory
        local_path = os.path.join(directory, "exploration")
        client = TarClient(local_path)
        self.assertFalse(client.path_exists())
        os.makedirs(local_path)
        self.assertTrue(client.checkout(self.remote_url))
        # non-empty
        self.assertFalse(client.checkout(self.remote_url))

    def test_checkout_version(self):
        directory = tempfile.mkdtemp()
        self.directories["checkout_test"] = directory
        local_path = os.path.join(directory, "exploration")
        client = TarClient(local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(self.remote_url,
                                        version=self.package_version))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), local_path)
        self.assertEqual(client.get_url(), self.remote_url)
        # make sure the tarball subdirectory was promoted correctly.
        self.assertTrue(os.path.exists(os.path.join(local_path, 'package.xml')))

    def test_get_environment_metadata(self):
        # Verify that metadata is generated
        directory = tempfile.mkdtemp()
        self.directories['local'] = directory
        local_path = os.path.join(directory, "local")
        client = TarClient(local_path)
        self.assertTrue('version' in client.get_environment_metadata())


class TarClientTestLocal(unittest.TestCase):
    '''Tests with URL being a local filepath'''

    def setUp(self):
        self.root_directory = tempfile.mkdtemp()
        # helpful when setting tearDown to pass
        self.directories = dict(setUp=self.root_directory)
        self.version_path0 = os.path.join(self.root_directory, "version")
        self.version_path1 = os.path.join(self.root_directory, "version1")
        self.version_path2 = os.path.join(self.root_directory, "version1.0")

        os.makedirs(self.version_path0)
        os.makedirs(self.version_path1)
        os.makedirs(self.version_path2)

        subprocess.check_call("touch stack0.xml", shell=True, cwd=self.version_path0)
        subprocess.check_call("touch stack.xml", shell=True, cwd=self.version_path1)
        subprocess.check_call("touch stack1.xml", shell=True, cwd=self.version_path2)
        subprocess.check_call("touch version1.txt", shell=True, cwd=self.root_directory)

        self.tar_url = os.path.join(self.root_directory, "origin.tar")
        self.tar_url_compressed = os.path.join(self.root_directory,
                                               "origin_compressed.tar.bz2")

        subprocess.check_call("tar -cf %s %s" % (self.tar_url, " ".join(["version",
                                                                        "version1",
                                                                        "version1.txt",
                                                                        "version1.0"])),
                              shell=True,
                              cwd=self.root_directory)
        subprocess.check_call("tar -cjf %s %s" % (self.tar_url_compressed, " ".join(["version",
                                                                                    "version1",
                                                                                    "version1.txt",
                                                                                    "version1.0"])),
                              shell=True,
                              cwd=self.root_directory)

    def tearDown(self):
        for d in self.directories:
            self.assertTrue(os.path.exists(self.directories[d]))
            shutil.rmtree(self.directories[d])
            self.assertFalse(os.path.exists(self.directories[d]))

    def test_checkout_version_local(self):
        directory = tempfile.mkdtemp()
        self.directories["checkout_test"] = directory
        local_path = os.path.join(directory, "version1")
        url = self.tar_url
        client = TarClient(local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url, version='version1'))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), local_path)
        self.assertEqual(client.get_url(), url)
        # make sure the tarball subdirectory was promoted correctly.
        self.assertTrue(os.path.exists(os.path.join(local_path, 'stack.xml')))

    def test_checkout_version_compressed_local(self):
        directory = tempfile.mkdtemp()
        self.directories["checkout_test"] = directory
        local_path = os.path.join(directory, "version1")
        url = self.tar_url_compressed
        client = TarClient(local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url, version='version1'))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), local_path)
        self.assertEqual(client.get_url(), url)
        # make sure the tarball subdirectory was promoted correctly.
        self.assertTrue(os.path.exists(os.path.join(local_path, 'stack.xml')))



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
"""
git vcs support.

refnames in git can be branchnames, hashes, partial hashes, tags. On
checkout, git will disambiguate by checking them in that order, taking
the first that applies

This class aims to provide git for linear centralized workflows. This
means in case of ambiguity, we assume that the only relevant remote
is the one named "origin", and we assume that commits once on origin
remain on origin.


A challenge with git is that it has strong reasonable conventions, but
is very allowing for breaking them. E.g. it is possible to name
remotes and branches with names like "refs/heads/master", give
branches and tags the same name, or a valid SHA-ID as name, etc.
Similarly git allows plenty of ways to reference any object, in case
of ambiguities, git attempts to take the most reasonable
disambiguation, and in some cases warns.
"""


from __future__ import absolute_import, print_function, unicode_literals
import os
import sys
import shutil
import tempfile
import gzip
import dateutil.parser  # For parsing date strings
from distutils.version import LooseVersion
import logging

from vcstools.vcs_base import VcsClientBase, VcsError
from vcstools.common import sanitized, normalized_rel_path, run_shell_command

from vcstools.git_archive_all import *


class GitError(Exception):
    pass


def _git_diff_path_submodule_change(diff, rel_path_prefix):
    """
    Parses git diff result and changes the filename prefixes.
    """
    if diff is None:
        return None
    INIT = 0
    INDIFF = 1
    # small state machine makes sure we never touch anything inside
    # the actual diff
    state = INIT
    result = ""
    s_list = [line for line in diff.split(os.linesep)]
    subrel_path = rel_path_prefix
    for line in s_list:
        newline = line
        if line.startswith("Entering '"):
            state = INIT
            submodulepath = line.rstrip("'")[len("Entering '"):]
            subrel_path = os.path.join(rel_path_prefix, submodulepath)
            continue
        if line.startswith("diff --git "):
            state = INIT
        if state == INIT:
            if line.startswith("@@"):
                state = INDIFF
            else:
                if line.startswith("---") and not line.startswith("--- /dev/null"):
                    newline = "--- " + subrel_path + line[5:]
                if line.startswith("+++") and not line.startswith("+++ /dev/null"):
                    newline = "+++ " + subrel_path + line[5:]
                if line.startswith("diff --git"):
                    # first replacing b in case path starts with a/
                    newline = line.replace(" b/", " " + subrel_path + "/", 1)
                    newline = newline.replace(" a/", " " + subrel_path + "/", 1)
        if newline != '':
            result += newline + '\n'
    return result


def _get_git_version():
    """Looks up git version by calling git --version.

    :raises: VcsError if git is not installed or returns
    something unexpected"""
    try:
        cmd = 'git --version'
        value, version, _ = run_shell_command(cmd, shell=True)
        if value != 0:
            raise VcsError("git --version returned %s, maybe git is not installed" % (value))
        prefix = 'git version '
        if version is not None and version.startswith(prefix):
            version = version[len(prefix):].strip()
        else:
            raise VcsError("git --version returned invalid string: '%s'" % version)
    except VcsError as exc:
        raise VcsError("Could not determine whether git is installed: %s" % exc)
    return version


class GitClient(VcsClientBase):
    def __init__(self, path):
        """
        :raises: VcsError if git not detected
        """
        VcsClientBase.__init__(self, 'git', path)
        self.gitversion = _get_git_version()

    @staticmethod
    def get_environment_metadata():
        metadict = {}
        try:
            version = _get_git_version()
            resetkeep = LooseVersion(version) >= LooseVersion('1.7.1')
            submodules = LooseVersion(version) > LooseVersion('1.7')
            metadict["features"] = "'reset --keep': %s, submodules: %s" % (resetkeep, submodules)
        except VcsError:
            version = "No git installed"
        metadict["version"] = version
        return metadict

    def get_url(self):
        """
        :returns: git URL of the directory path (output of git info command), or None if it cannot be determined
        """
        if self.detect_presence():
            cmd = "git config --get remote.%s.url" % self._get_default_remote()
            _, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
            return output.rstrip()
        return None

    def _get_default_remote(self):
        """
        in order to support users who name their default origin
        something else than origin, read remote name.
        """
        # TODO: maybe pick other remote depending on context
        return 'origin'

    @staticmethod
    def static_detect_presence(path):
        # There is a proposed implementation of detect_presence which might be
        # more future proof, but would depend on parsing the output of git
        # See: https://github.com/vcstools/vcstools/pull/10
        return os.path.exists(os.path.join(path, '.git'))

    def checkout(self, url, version=None, verbose=False, shallow=False, timeout=None):
        """calls git clone and then, if version was given, update(version)"""
        if url is None or url.strip() == '':
            raise ValueError('Invalid empty url : "%s"' % url)

        # since we cannot know whether version names a branch, clone master initially
        cmd = 'git clone'
        if shallow:
            cmd += ' --depth 1'
            if LooseVersion(self.gitversion) >= LooseVersion('1.7.10'):
                cmd += ' --no-single-branch'
        if version is None:
            # quicker than using _do_update, but undesired when switching branches next
            cmd += ' --recursive'
        cmd += ' %s %s' % (url, self._path)
        value, _, msg = run_shell_command(cmd,
                                          shell=True,
                                          no_filter=True,
                                          show_stdout=verbose,
                                          timeout=timeout,
                                          verbose=verbose)
        if value != 0:
            if msg:
                self.logger.error('%s' % msg)
            return False

        try:
            # update to make sure we are on the right branch. Do not
            # check for "master" here, as default branch could be anything
            if version is not None:
                return self._do_update(version,
                                       verbose=verbose,
                                       fast_foward=True,
                                       timeout=timeout,
                                       update_submodules=True)
            else:
                return True
        except GitError:
            return False

    def _update_submodules(self, verbose=False, timeout=None):

        # update submodules ( and init if necessary ).
        if LooseVersion(self.gitversion) > LooseVersion('1.7'):
            cmd = "git submodule update --init --recursive"
            value, _, _ = run_shell_command(cmd,
                                            shell=True,
                                            cwd=self._path,
                                            show_stdout=True,
                                            timeout=timeout,
                                            verbose=verbose)
            if value != 0:
                return False
        return True

    def update(self, version=None, verbose=False, force_fetch=False, timeout=None):
        """
        if version is None, attempts fast-forwarding current branch, if any.

        Else interprets version as a local branch, remote branch, tagname,
        hash, etc.

        If it is a branch, attempts to move to it unless
        already on it, and to fast-forward, unless not a tracking
        branch. Else go untracked on tag or whatever version is. Does
        not leave if current commit would become dangling.

        :return: True if already up-to-date with remote or after successful fast_foward
        """
        if not self.detect_presence():
            return False

        try:
            # fetch in any case to get updated tags even if we don't need them
            self._do_fetch()
            return self._do_update(refname=version, verbose=verbose, timeout=timeout)
        except GitError:
            return False

    def _do_update(self,
                   refname=None,
                   verbose=False,
                   fast_foward=True,
                   timeout=None,
                   update_submodules=True):
        '''
        updates without fetching, thus any necessary fetching must be done before
        allows arguments to reduce unnecessary steps after checkout

        :param fast_foward: if false, does not perform fast-forward
        :param update_submodules: if false, does not attempt to update submodules
        '''
        # are we on any branch?
        current_branch = self._get_branch()
        branch_parent = None
        if current_branch:
            # local branch might be named differently from remote by user, we respect that
            same_branch = (refname == current_branch)
            if not same_branch:
                (branch_parent, remote) = self._get_branch_parent(current_branch=current_branch)
                if not refname:
                    # ! changing refname to cause fast-forward
                    refname = branch_parent
                    same_branch = True
                else:
                    same_branch = (refname == branch_parent)
                if same_branch and not branch_parent:
                    # avoid expensive checking branch parent again later
                    fast_foward = False
        else:
            same_branch = False

        if not refname:
            # we are neither tracking, nor did we get any refname to update to
            return (not update_submodules) or self._update_submodules(verbose=verbose,
                                                                      timeout=timeout)

        default_remote = self._get_default_remote()
        if same_branch:
            if fast_foward:
                if not branch_parent and current_branch:
                    (branch_parent, remote) = self._get_branch_parent(current_branch=current_branch)
                    if remote != default_remote:
                        # if remote is not origin, must not fast-forward (because based on origin)
                        logger = logging.getLogger('vcstools')
                        logger.warn("vcstools only handles branches tracking default remote,"
                                    " branch '%s' tracks remote '%s'.\nRepository path is '%s'."
                                    % (current_branch, remote, self._path))
                        branch_parent = None
                # already on correct branch, fast-forward if there is a parent
                if branch_parent:
                    if not self._do_fast_forward(branch_parent=branch_parent,
                                                 fetch=False,
                                                 verbose=verbose):
                        return False
        else:
            # refname can be a different branch or something else than a branch

            refname_is_local_branch = self._is_local_branch(refname)
            if refname_is_local_branch:
                # might also be remote branch, but we treat it as local
                refname_is_remote_branch = False
            else:
                refname_is_remote_branch = self._is_remote_branch(refname, fetch=False)
            refname_is_branch = refname_is_remote_branch or refname_is_local_branch

            current_version = None
            # shortcut if version is the same as requested
            if not refname_is_branch:
                current_version = self.get_version()
                if current_version == refname:
                    return (not update_submodules) or self._update_submodules(verbose=verbose,
                                                                              timeout=timeout)

            if current_branch is None:
                if not current_version:
                    current_version = self.get_version()
                # prevent commit from becoming dangling
                if self._is_commit_in_orphaned_subtree(current_version, fetch=False):
                    # commit becomes dangling unless we move to one of its descendants
                    if not self._rev_list_contains(refname, current_version, fetch=False):
                        # TODO: should raise error instead of printing message
                        sys.stderr.write("vcstools refusing to move away from dangling commit, to protect your work.\n")
                        return False

            # git checkout makes all the decisions for us
            self._do_checkout(refname, verbose=verbose, fetch=False)

            if refname_is_local_branch:
                # if we just switched to a local tracking branch (not created one), we should also fast forward
                (new_branch_parent, remote) = self._get_branch_parent(current_branch=refname)
                if remote != default_remote:
                    # if remote is not origin, must not fast-forward (because based on origin)
                    sys.stderr.write("vcstools only handles branches tracking default remote," +
                                     " branch '%s' tracks remote '%s'\n" % (current_branch, remote))
                    new_branch_parent = None
                if new_branch_parent is not None:
                    if fast_foward:
                        if not self._do_fast_forward(branch_parent=new_branch_parent,
                                                     fetch=False,
                                                     verbose=verbose):
                            return False
        return (not update_submodules) or self._update_submodules(verbose=verbose, timeout=timeout)

    def get_current_version_label(self):
        """
        For git we change the label to clarify when a different remote
        is configured.
        """
        branch = self._get_branch()
        if branch is None:
            return '<detached>'
        result = branch
        (remote_branch, remote) = self._get_branch_parent()
        if remote_branch is not None:
            default_remote = self._get_default_remote()
            # if not following 'origin/branch', display 'branch < tracked ref'
            if (remote_branch != branch or remote != default_remote):
                result += ' < '
                if remote != default_remote:
                    result += remote + '/'
                result += remote_branch
        return result

    def get_default_remote_version_label(self):
        if self.detect_presence():
            _, output, _ = run_shell_command('git remote show %s' % self._get_default_remote(),
                                             shell=True,
                                             cwd=self._path)
            for line in output.splitlines():
                elems = line.split()
                if elems[0:2] == ['HEAD', 'branch:']:
                    return elems[2]
        return None

    def get_remote_version(self, fetch=False):
        # try tracked branch on origin (returns None if on other remote)
        (parent_branch, remote) = self._get_branch_parent(fetch=fetch)
        if parent_branch is not None:
            return self.get_version(spec=remote+'/'+parent_branch)

    def get_version(self, spec=None):
        """
        :param spec: (optional) token to identify desired version. For
          git, this may be anything accepted by git log, e.g. a tagname,
          branchname, or sha-id.
        :param fetch: When spec is given, can be used to suppress git fetch call
        :returns: current SHA-ID of the repository. Or if spec is
          provided, the SHA-ID of a commit specified by some token if found, else None
        """
        if self.detect_presence():
            command = "git log -1"
            if spec is not None:
                command += " %s" % sanitized(spec)
            command += " --format='%H'"
            _, output, _ = run_shell_command(command, shell=True,
                                             no_warn=True, cwd=self._path)
            if output.strip() != '':
                # On Windows the version can have single quotes around it
                version = output.strip().strip("'")
                return version  # found SHA-ID
            elif spec is None:
                return None
            # we try again after fetching if given spec had not been found
            try:
                self._do_fetch()
            except GitError:
                return None
            # we repeat the call once again after fetching
            _, output, _ = run_shell_command(command, shell=True,
                                             no_warn=True, cwd=self._path)
            if output.strip() == '':
                # even if after fetching, not found specified version
                return None
            version = output.strip().strip("'")
            return version
        return None

    def get_diff(self, basepath=None):
        response = ''
        if basepath is None:
            basepath = self._path
        if self.path_exists():
            rel_path = normalized_rel_path(self._path, basepath)
            # git needs special treatment as it only works from inside
            # use HEAD to also show staged changes. Maybe should be option?
            # injection should be impossible using relpath, but to be sure, we check
            cmd = "git diff HEAD --src-prefix=%s/ --dst-prefix=%s/ ." % \
                  (sanitized(rel_path), sanitized(rel_path))
            _, response, _ = run_shell_command(cmd, shell=True, cwd=self._path)
            if LooseVersion(self.gitversion) > LooseVersion('1.7'):
                cmd = 'git submodule foreach --recursive git diff HEAD'
                _, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
                response += _git_diff_path_submodule_change(output, rel_path)
        return response

    def get_affected_files(self, revision):
        # Making changes for windows support
        cmd = "git show {0} --pretty=format: --name-only".format(
            revision)
        code, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
        affected = []
        if code == 0:
            for filename in output.splitlines():
                if filename not in ('', None, ):
                    affected.append(filename)
        return affected

    def get_log(self, relpath=None, limit=None):
        response = []

        if relpath is None:
            relpath = ''

        if self.path_exists() and os.path.exists(os.path.join(self._path, relpath)):
            # Get the log
            limit_cmd = (("-n %d" % (int(limit))) if limit else "")

            GIT_COMMIT_FIELDS = ['id', 'author', 'email', 'date', 'message']
            GIT_LOG_FORMAT = '%x1f'.join(['%H', '%an', '%ae', '%ad', '%s']) + '%x1e'

            command = "git --work-tree=%s log --format=\"%s\" %s %s " % (self._path, GIT_LOG_FORMAT,
                                                                         limit_cmd, sanitized(relpath))
            return_code, response_str, stderr = run_shell_command(command, shell=True, cwd=self._path)

            if return_code == 0:
                # Parse response
                response = response_str.strip('\n\x1e').split("\x1e")
                response = [row.strip().split("\x1f") for row in response]
                response = [dict(zip(GIT_COMMIT_FIELDS, row)) for row in response]

                # Parse dates
                for entry in response:
                    entry['date'] = dateutil.parser.parse(entry['date'])

        return response

    def get_status(self, basepath=None, untracked=False, porcelain=False):
        status_flag = '--porcelain' if porcelain else '-s'
        response = None
        if basepath is None:
            basepath = self._path
        if self.path_exists():
            rel_path = normalized_rel_path(self._path, basepath)
            # git command only works inside repo
            # self._path is safe against command injection, as long as we check path.exists
            command = "git status {0} ".format(status_flag)
            if not untracked:
                command += " -uno"
            _, response, _ = run_shell_command(command,
                                               shell=True,
                                               cwd=self._path)
            response_processed = ""
            for line in response.split('\n'):
                if len(line.strip()) > 0:
                    # prepend relative path
                    response_processed += '%s%s/%s\n' % (line[0:3],
                                                         rel_path,
                                                         line[3:])
            if LooseVersion(self.gitversion) > LooseVersion('1.7'):
                if not untracked:
                    status_flag += " -uno"
                command = "git submodule foreach --recursive 'git status {0}'".format(status_flag)
                _, response2, _ = run_shell_command(command,
                                                    shell=True,
                                                    cwd=self._path)
                for line in response2.split('\n'):
                    if line.startswith("Entering"):
                        continue
                    if len(line.strip()) > 0:
                        # prepend relative path
                        response_processed += line[0:3] + rel_path + '/' + line[3:] + '\n'
            response = response_processed
        return response

    def _is_remote_branch(self, branch_name, remote_name=None, fetch=True):
        """
        checks list of remote branches for match. Set fetch to False if you just fetched already.

        :returns: True if branch_name exists for remote <remote_name> (or 'origin' if None)
        :raises: GitError when git fetch fails
        """
        if remote_name is None:
            remote_name = self._get_default_remote()

        if self.path_exists():
            if fetch:
                self._do_fetch()
            _, output, _ = run_shell_command('git branch -r',
                                             shell=True,
                                             cwd=self._path)
            for l in output.splitlines():
                elem = l.split()[0]
                rem_name = elem[:elem.find('/')]
                br_name = elem[elem.find('/') + 1:]
                if rem_name == remote_name and br_name == branch_name:
                    return True
        return False

    def _is_local_branch(self, branch_name):
        if self.path_exists():
            _, output, _ = run_shell_command('git branch',
                                             shell=True,
                                             cwd=self._path)
            for line in output.splitlines():
                elems = line.split()
                if len(elems) == 1:
                    if elems[0] == branch_name:
                        return True
                elif len(elems) == 2:
                    if elems[0] == '*' and elems[1] == branch_name:
                        return True
        return False

    def _get_branch(self):
        if self.path_exists():
            _, output, _ = run_shell_command('git branch',
                                             shell=True,
                                             cwd=self._path)
            for line in output.splitlines():
                elems = line.split()
                if len(elems) == 2 and elems[0] == '*':
                    return elems[1]
        return None

    def _get_branch_parent(self, fetch=False, current_branch=None):
        """
        :param fetch: if true, performs git fetch first
        :param current_branch: if not None, this is used as current branch (else extra shell call)
        :returns: (branch, remote) the name of the branch this branch tracks and its remote
        :raises: GitError if fetch fails
        """
        if not self.path_exists():
            return (None, None)
        # get name of configured merge ref.
        branchname = current_branch or self._get_branch()
        if branchname is None:
            return (None, None)

        cmd = 'git config --get %s' % sanitized('branch.%s.merge' % branchname)

        _, output, _ = run_shell_command(cmd,
                                         shell=True,
                                         cwd=self._path)
        if not output:
            return (None, None)
        lines = output.splitlines()
        if len(lines) > 1:
            sys.stderr.write("vcstools unable to handle multiple merge references for branch %s:\n%s\n"
                             % (branchname, output))
            return (None, None)

        # get name of configured remote
        cmd = 'git config --get "branch.%s.remote"' % branchname
        _, output2, _ = run_shell_command(cmd, shell=True, cwd=self._path)
        remote = output2 or self._get_default_remote()

        branch_reference = lines[0]
        # branch_reference is either refname, or /refs/heads/refname, or
        # heads/refname we would like to return refname however,
        # user could also have named any branch
        # "/refs/heads/refname", for some unholy reason check all
        # known branches on remote for refname, then for the odd
        # cases, as git seems to do
        candidate = branch_reference
        if candidate.startswith('refs/'):
            candidate = candidate[len('refs/'):]
        if candidate.startswith('heads/'):
            candidate = candidate[len('heads/'):]
        elif candidate.startswith('tags/'):
            candidate = candidate[len('tags/'):]
        elif candidate.startswith('remotes/'):
            candidate = candidate[len('remotes/'):]

        result = None
        if self._is_remote_branch(candidate, remote_name=remote, fetch=fetch):
            result = candidate
        elif branch_reference != candidate and self._is_remote_branch(branch_reference,
                                                                      remote_name=remote,
                                                                      fetch=False):
            result = branch_reference

        if result is not None:
            return (result, remote)
        return None, None

    def is_tag(self, tag_name, fetch=True):
        """
        checks list of tags for match.
        Set fetch to False if you just fetched already.

        :returns: True if tag_name among known tags
        :raises: GitError when call to git fetch fails
        """
        if fetch:
            self._do_fetch()
        if not tag_name:
            raise ValueError('is_tag requires tag_name, got: "%s"' % tag_name)
        if self.path_exists():
            cmd = 'git tag -l %s' % sanitized(tag_name)
            _, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
            lines = output.splitlines()
            if len(lines) == 1:
                return True
        return False

    def _rev_list_contains(self, refname, version, fetch=True):
        """
        calls git rev-list with refname and returns True if version
        can be found in rev-list result

        :param refname: a git refname
        :param version: an SHA IDs (if partial, caller is responsible
          for mismatch)
        :returns: True if version is an ancestor commit from refname
        :raises: GitError when call to git fetch fails
        """
        # to avoid listing unnecessarily many rev-ids, we cut off all
        # those we are definitely not interested in
        # $ git rev-list foo bar ^baz ^bez
        # means "list all the commits which are reachable from foo or
        # bar, but not from baz or bez". We use --parents because
        # ^baz also excludes baz itself. We could also use git
        # show --format=%P to get all parents first and use that,
        # not sure what's more performant
        if fetch:
            self._do_fetch()
        if (refname is not None and refname != '' and
                version is not None and version != ''):

            cmd = 'git rev-list %s ^%s --parents' % (sanitized(refname), sanitized(version))
            _, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
            for line in output.splitlines():
                # can have 1, 2 or 3 elements (commit, parent1, parent2)
                for hashid in line.split(" "):
                    if hashid.startswith(version):
                        return True
        return False

    def _is_commit_in_orphaned_subtree(self, version, mask_self=False, fetch=True):
        """
        checks git log --all (the list of all commits reached by
        references, meaning branches or tags) for version. If it shows
        up, that means git garbage collection will not remove the
        commit. Else it would eventually be deleted.

        :param version: SHA IDs (if partial, caller is responsible for mismatch)
        :param mask_self: whether to consider direct references to this commit
            (rather than only references on descendants) as well
        :param fetch: whether fetch should be done first for remote refs
        :returns: True if version is not recursively referenced by a branch or tag
        :raises: GitError if git fetch fails
        """
        if fetch:
            self._do_fetch()
        if version is not None and version != '':
            cmd = 'git show-ref -s'
            _, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
            refs = output.splitlines()
            # 2000 seems like a number the linux shell can cope with
            chunksize = 2000
            refchunks = [refs[x:x + chunksize] for x in range(0, len(refs), chunksize)]
            for refchunk in refchunks:
                # git log over all refs except HEAD
                cmd = 'git log ' + " ".join(refchunk)
                if mask_self:
                    # %P: parent hashes
                    cmd += " --pretty=format:%P"
                else:
                    # %H: commit hash
                    cmd += " --pretty=format:%H"
                _, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
                for line in output.splitlines():
                    if line.strip("'").startswith(version):
                        return False
            return True
        return False

    def export_repository(self, version, basepath):
        if not self.detect_presence():
            return False

        # Export should work regardless of the state of the local dir (might have modified files),
        # so we clone to a temporary folder and checkout the specified version there.
        # If the repo has submodules with relative URLs, cloning to a temp dir doesn't work.
        # To support this case, first check if current version is already checked out and there are no
        # modifications and try to export from current local dir.
        # see https://github.com/vcstools/vcstools/pull/130
        current_sha = self.get_version()
        export_sha = self.get_version(version)
        if current_sha == export_sha and self.get_diff() == '':
            archiver = GitArchiver(main_repo_abspath=self.get_path(), force_sub=True)
            filepath = '{0}.tar.gz'.format(basepath)
            archiver.create(filepath)
            return filepath

        # export a different version than currently checked out or there are local changes
        try:
            # since version may relate to remote branch / tag we do not
            # know about yet, do fetch if not already done
            self._do_fetch()
            # export should work regardless of the state of the local dir (might have modified files)
            # so we clone to a temporary folder and checkout the specified version there
            tmpd_path = tempfile.mkdtemp()
            try:
                tmpgit = GitClient(tmpd_path)
                if tmpgit.checkout(self._path, version=version, shallow=True):
                    archiver = GitArchiver(main_repo_abspath=tmpgit.get_path(), force_sub=True)
                    filepath = '{0}.tar.gz'.format(basepath)
                    archiver.create(filepath)
                    return filepath
                else:
                    return False
            finally:
                shutil.rmtree(tmpd_path)

        except GitError:
            return False

    def get_branches(self, local_only=False):
        cmd = 'git branch --no-color'
        if not local_only:
            cmd += ' -a'
        result, out, err = run_shell_command(cmd,
                                             cwd=self._path,
                                             shell=True,
                                             show_stdout=False)
        branches = []
        for line in out.splitlines():
            if 'HEAD -> ' in line:
                continue
            line = line.strip('* ')
            branches.append(line)
        return branches

    def _do_fetch(self, timeout=None):
        """
        calls git fetch
        :raises: GitError when call fails
        """
        cmd = "git fetch"
        value1, _, _ = run_shell_command(cmd,
                                         cwd=self._path,
                                         shell=True,
                                         no_filter=True,
                                         timeout=timeout,
                                         show_stdout=True)
        # git fetch --tags ONLY fetches new tags and commits used, no other commits!
        cmd = "git fetch --tags"
        value2, _, _ = run_shell_command(cmd,
                                         cwd=self._path,
                                         shell=True,
                                         no_filter=True,
                                         timeout=timeout,
                                         show_stdout=True)
        if value1 != 0 or value2 != 0:
            raise GitError('git fetch failed')

    def _do_fast_forward(self, branch_parent, fetch=True, verbose=False):
        """Execute git fetch if necessary, and if we can fast-foward,
        do so to the last fetched version using git rebase.

        :param branch_parent: name of branch we track
        :param fetch: whether fetch should be done first for remote refs
        :returns: True if up-to-date or after succesful fast-forward
        :raises: GitError when git fetch fails
        """
        assert branch_parent is not None
        current_version = self.get_version()
        default_remote = self._get_default_remote()
        parent_version = self.get_version("remotes/%s/%s" % (default_remote, branch_parent))
        if current_version == parent_version:
            return True
        # check if we are true ancestor of tracked branch
        if not self._rev_list_contains(parent_version,
                                       current_version,
                                       fetch=fetch):
            # if not rev_list_contains this version, we are on same
            # commit (checked before), have advanced, or have diverged.
            # Now check whether tracked branch is a true ancestor of us
            if self._rev_list_contains(current_version,
                                       parent_version,
                                       fetch=False):
                return True
            print("Cannot fast-forward, local repository and remote '%s' have diverged." % branch_parent)
            return False
        # 'git reset --keep' doesn't refresh the index. Do it manually to avoid
        # errors as reported in: https://github.com/vcstools/wstool/issues/77
        cmd = "git update-index -q --refresh"
        run_shell_command(cmd,
                          shell=True,
                          cwd=self._path,
                          show_stdout=False,
                          verbose=verbose)
        if verbose:
            print("Rebasing repository")
        # Rebase, do not pull, because somebody could have
        # commited in the meantime.
        if LooseVersion(self.gitversion) >= LooseVersion('1.7.1'):
            # --keep allows to rebase even with local changes, as long as
            # local changes are not in files that change between versions
            cmd = "git reset --keep remotes/%s/%s" % (default_remote, branch_parent)
            value, _, _ = run_shell_command(cmd,
                                            shell=True,
                                            cwd=self._path,
                                            show_stdout=True,
                                            verbose=verbose)
            if value == 0:
                return True
        else:
            verboseflag = ''
            if verbose:
                verboseflag = '-v'
            # prior to version 1.7.1, git does not know --keep
            # Do not merge, rebase does nothing when there are local changes
            cmd = "git rebase %s remotes/%s/%s" % (verboseflag, default_remote, branch_parent)
            value, _, _ = run_shell_command(cmd,
                                            shell=True,
                                            cwd=self._path,
                                            show_stdout=True,
                                            verbose=verbose)
            if value == 0:
                return True
        return False

    def _do_checkout(self, refname, fetch=True, verbose=False):
        """
        meaning git checkout, not vcstools checkout. This works
        for local branches, remote branches, tagnames, hashes, etc.
        git will create local branch of same name when no such local
        branch exists, and also setup tracking. Git decides with own
        rules whether local changes would cause conflicts, and refuses
        to checkout else.

        :raises GitError: when checkout fails
        """
        # since refname may relate to remote branch / tag we do not
        # know about yet, do fetch if not already done
        if fetch:
            self._do_fetch()
        cmd = "git checkout %s" % (refname)
        value, _, _ = run_shell_command(cmd,
                                        shell=True,
                                        cwd=self._path,
                                        show_stdout=verbose,
                                        verbose=verbose)
        if value != 0:
            raise GitError('Git Checkout failed')


# Backwards compatibility
GITClient = GitClient



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, print_function, unicode_literals
import os
import warnings


_VCS_TYPES = {}


def register_vcs(vcs_type, clazz):
    """
    :param vcs_type: id, ``str``
    :param clazz: class extending VcsClientBase
    """
    _VCS_TYPES[vcs_type] = clazz


def get_registered_vcs_types():
    """
    :returns: list of valid key to use as vcs_type
    """
    return list(_VCS_TYPES.keys())


def get_vcs(vcs_type):
    """
    Returns the class interfacing with vcs of given type

    :param vcs_type: id of the tpye, e.g. git, svn, hg, bzr
    :returns: class extending VcsClientBase
    :raises: ValueError for unknown vcs_type
    """
    vcs_class = _VCS_TYPES.get(vcs_type, None)
    if not vcs_class:
        raise ValueError('No Client type registered for vcs type "%s"' % vcs_type)
    return vcs_class


def get_vcs_client(vcs_type, path):
    """
    Returns a client with which to interact with the vcs at given path

    :param vcs_type: id of the tpye, e.g. git, svn, hg, bzr
    :returns: instance of VcsClientBase
    :raises: ValueError for unknown vcs_type
    """
    clientclass = get_vcs(vcs_type)
    return clientclass(path)


class VcsClient(object):
    """
    *DEPRECATED* API for interacting with source-controlled paths
    independent of actual version-control implementation.
    """

    def __init__(self, vcs_type, path):
        self._path = path
        warnings.warn("Class VcsClient is deprecated, use from vcstools" +
                      " import get_vcs_client; get_vcs_client() instead")
        self.vcs = get_vcs_client(vcs_type, path)

    def path_exists(self):
        return os.path.exists(self._path)

    def get_path(self):
        return self._path

    # pass through VCSClientBase API
    def get_version(self, spec=None):
        return self.vcs.get_version(spec)

    def get_current_version_label(self):
        return self.vcs.get_current_version_label()

    def get_remote_version(self, fetch=False):
        return self.vcs.get_remote_version(fetch)

    def get_default_remote_version_label(self):
        return self.vcs.get_default_remote_version_label()

    def checkout(self, url, version='', verbose=False, shallow=False):
        return self.vcs.checkout(url,
                                 version,
                                 verbose=verbose,
                                 shallow=shallow)

    def url_matches(self, url, url_or_shortcut):
        return self.vcs.url_matches(url=url, url_or_shortcut=url_or_shortcut)

    def update(self, version='', verbose=False):
        return self.vcs.update(version, verbose=verbose)

    def detect_presence(self):
        return self.vcs.detect_presence()

    def get_vcs_type_name(self):
        return self.vcs.get_vcs_type_name()

    def get_url(self):
        return self.vcs.get_url()

    def get_diff(self, basepath=None):
        return self.vcs.get_diff(basepath)

    def get_status(self, basepath=None, untracked=False, **kwargs):
        return self.vcs.get_status(basepath, untracked, **kwargs)

    def get_log(self, relpath=None, limit=None):
        return self.vcs.get_log(relpath, limit)

    def export_repository(self, version, basepath):
        return self.vcs.export_repository(version, basepath)

    def get_branches(self, local_only=False):
        return self.vcs.get_branches(local_only)


# backwards compat
VCSClient = VcsClient



#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, print_function, unicode_literals

import os
import io
import unittest
import subprocess
import tempfile
import shutil
import re
from vcstools.svn import SvnClient, canonical_svn_url_split, get_remote_contents


class SvnClientUtilTest(unittest.TestCase):

    def test_canonical_svn_url_split(self):
        self.assertEqual({'root': 'foo',
                          'type': None,
                          'name': None, 'subfolder': None,
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('foo'))
        self.assertEqual({'root': None,
                          'type': None,
                          'name': None, 'subfolder': None,
                          'query': None, 'fragment': None},
                         canonical_svn_url_split(None))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'branches',
                          'name': 'foo', 'subfolder': None,
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/branches/foo'))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'branches',
                          'name': 'foo', 'subfolder': None,
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/branches/foo/'))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'branches',
                          'name': 'foo', 'subfolder': 'sub/bar',
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/branches/foo/sub/bar'))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'trunk',
                          'name': None, 'subfolder': None,
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/trunk'))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'trunk',
                          'name': None, 'subfolder': 'sub',
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/trunk/sub'))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'trunk',
                          'name': None, 'subfolder': 'sub/foo',
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/trunk/sub/foo'))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'tags',
                          'name': '1.2.3', 'subfolder': None,
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/tags/1.2.3'))
        self.assertEqual({'root': 'svn://gcc.gnu.org/svn/gcc',
                          'type': 'tags',
                          'name': '1.2.3', 'subfolder': 'sub/foo',
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('svn://gcc.gnu.org/svn/gcc/tags/1.2.3/sub/foo'))
        self.assertEqual({'root': 'file://localhost/svn/gcc',
                          'type': 'tags',
                          'name': '1.2.3', 'subfolder': 'sub/foo',
                          'query': None, 'fragment': None},
                         canonical_svn_url_split('file://localhost/svn/gcc/tags/1.2.3/sub/foo'))
        self.assertEqual({'root': 'https://frodo@gcc.gnu.org/svn/gcc',
                          'type': 'tags',
                          'name': '1.2.3', 'subfolder': 'sub/foo',
                          'query': 'pw=guest', 'fragment': 'today'},
                         canonical_svn_url_split('https://frodo@gcc.gnu.org/svn/gcc/tags/1.2.3/sub/foo?pw=guest#today'))


class SvnClientTestSetups(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.root_directory = tempfile.mkdtemp()
        self.directories = dict(setUp=self.root_directory)
        self.remote_path = os.path.join(self.root_directory, "remote")
        self.init_path = os.path.join(self.root_directory, "init")

        # create a "remote" repo
        subprocess.check_call("svnadmin create %s" % self.remote_path, shell=True, cwd=self.root_directory)
        self.local_root_url = "file://localhost" + self.remote_path
        self.local_url = self.local_root_url + "/trunk"

        # create an "init" repo to populate remote repo
        subprocess.check_call("svn checkout %s %s" % (self.local_root_url, self.init_path), shell=True, cwd=self.root_directory)

        for cmd in [
            "mkdir trunk",
            "mkdir branches",
            "mkdir tags",
            "svn add trunk branches tags",
            "touch trunk/fixed.txt",
            "svn add trunk/fixed.txt",
            "svn commit -m initial"]:
            subprocess.check_call(cmd, shell=True, cwd=self.init_path)

        self.local_version_init = "-r1"

        # files to be modified in "local" repo
        for cmd in [
            "touch trunk/modified.txt",
            "touch trunk/modified-fs.txt",
            "svn add trunk/modified.txt trunk/modified-fs.txt",
            "svn commit -m initial"]:
            subprocess.check_call(cmd, shell=True, cwd=self.init_path)

        self.local_version_second = "-r2"
        for cmd in [
            "touch trunk/deleted.txt",
            "touch trunk/deleted-fs.txt",
            "svn add trunk/deleted.txt trunk/deleted-fs.txt",
            "svn commit -m modified"]:
            subprocess.check_call(cmd, shell=True, cwd=self.init_path)

        self.local_version_master = "-r3"

        # files to be modified in "local" repo
        for cmd in [
            "mkdir branches/foo",
            "touch branches/foo/modified.txt",
            "svn add branches/foo",
            "svn commit -m 'foo branch'"]:
            subprocess.check_call(cmd, shell=True, cwd=self.init_path)
        self.branch_url = self.local_root_url + "/branches/foo"
        self.local_version_foo_branch = "-r4"

        self.local_path = os.path.join(self.root_directory, "local")

    @classmethod
    def tearDownClass(self):
        for d in self.directories:
            shutil.rmtree(self.directories[d])

    def tearDown(self):
        if os.path.exists(self.local_path):
            shutil.rmtree(self.local_path)


class SvnClientTest(SvnClientTestSetups):

    def test_get_url_by_reading(self):
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(self.local_url, client.get_url())
        self.assertEqual(client.get_version(), self.local_version_master)
        self.assertEqual(client.get_version("PREV"), "-r2")
        self.assertEqual(client.get_version("2"), "-r2")
        self.assertEqual(client.get_version("-r2"), "-r2")
        # test invalid cient and repo without url
        client = SvnClient(os.path.join(self.remote_path, 'foo'))
        self.assertEqual(None, client.get_url())

    def test_get_type_name(self):
        local_path = "/tmp/dummy"
        client = SvnClient(local_path)
        self.assertEqual(client.get_vcs_type_name(), 'svn')

    def test_get_url_nonexistant(self):
        local_path = "/tmp/dummy"
        client = SvnClient(local_path)
        self.assertEqual(client.get_url(), None)

    def test_checkout(self):
        url = self.local_url
        client = SvnClient(self.local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), self.local_path)
        self.assertEqual(client.get_url(), url)

    def test_checkout_dir_exists(self):
        url = self.local_url
        client = SvnClient(self.local_path)
        self.assertFalse(client.path_exists())
        os.makedirs(self.local_path)
        self.assertTrue(client.checkout(url))
        # non-empty
        self.assertFalse(client.checkout(url))

    def test_checkout_emptyversion(self):
        url = self.local_url
        client = SvnClient(self.local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url, version=''))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), self.local_path)
        self.assertEqual(client.get_url(), url)
        self.assertTrue(client.update(None))
        self.assertTrue(client.update(""))

    def test_checkout_specific_version_and_update_short(self):
        "using just a number as version"
        url = self.local_url
        version = "3"
        client = SvnClient(self.local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url, version))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_version(), "-r3")
        new_version = '2'
        self.assertTrue(client.update(new_version))
        self.assertEqual(client.get_version(), "-r2")

    def test_get_remote_version(self):
        url = self.local_url
        client = SvnClient(self.local_path)
        client.checkout(url)
        self.assertEqual(client.get_remote_version(fetch=True),
                         self.local_version_master)
        self.assertEqual(client.get_remote_version(fetch=False),
                         None)

    def test_get_remote_branch_version(self):
        url = self.branch_url
        client = SvnClient(self.local_path)
        client.checkout(url)
        self.assertEqual(client.get_remote_version(fetch=True),
                         self.local_version_foo_branch)
        self.assertEqual(client.get_remote_version(fetch=False),
                         None)


    def testDiffClean(self):
        client = SvnClient(self.remote_path)
        self.assertEquals('', client.get_diff())

    def testStatusClean(self):
        client = SvnClient(self.remote_path)
        self.assertEquals('', client.get_status())

    def test_get_environment_metadata(self):
        # Verify that metadata is generated
        directory = tempfile.mkdtemp()
        self.directories['local'] = directory
        local_path = os.path.join(directory, "local")
        client = SvnClient(local_path)
        self.assertTrue('version' in client.get_environment_metadata())


class SvnClientLogTest(SvnClientTestSetups):

    @classmethod
    def setUpClass(self):
        SvnClientTestSetups.setUpClass()
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)

    def test_get_log_defaults(self):
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log()
        self.assertEquals(3, len(log))
        self.assertEquals('modified', log[0]['message'])
        for key in ['id', 'author', 'date', 'message']:
            self.assertTrue(log[0][key] is not None, key)
        # svn logs don't have email, but key should be in dict
        self.assertTrue(log[0]['email'] is None)

    def test_get_log_limit(self):
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log(limit=1)
        self.assertEquals(1, len(log))
        self.assertEquals('modified', log[0]['message'])

    def test_get_log_path(self):
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log(relpath='fixed.txt')
        self.assertEquals('initial', log[0]['message'])


class SVNClientAffectedFiles(SvnClientTestSetups):

    @classmethod
    def setUpClass(self):
        SvnClientTestSetups.setUpClass()
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)

    def test_get_affected_files(self):
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log(limit=1)[0]
        affected = client.get_affected_files(log['id'])

        self.assertEqual(sorted(['deleted-fs.txt', 'deleted.txt']),
                         sorted(affected))


class SvnDiffStatClientTest(SvnClientTestSetups):

    @classmethod
    def setUpClass(self):
        SvnClientTestSetups.setUpClass()
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)
        # after setting up "local" repo, change files and make some changes
        subprocess.check_call("rm deleted-fs.txt", shell=True, cwd=self.local_path)
        subprocess.check_call("svn rm deleted.txt", shell=True, cwd=self.local_path)
        f = io.open(os.path.join(self.local_path, "modified.txt"), 'a')
        f.write('0123456789abcdef')
        f.close()
        f = io.open(os.path.join(self.local_path, "modified-fs.txt"), 'a')
        f.write('0123456789abcdef')
        f.close()
        f = io.open(os.path.join(self.local_path, "added-fs.txt"), 'w')
        f.write('0123456789abcdef')
        f.close()
        f = io.open(os.path.join(self.local_path, "added.txt"), 'w')
        f.write('0123456789abcdef')
        f.close()
        subprocess.check_call("svn add added.txt", shell=True, cwd=self.local_path)

    def tearDown(self):
        pass

    def assertStatusListEqual(self, listexpect, listactual):
        """helper fun to check scm status output while discarding file ordering differences"""
        lines_expect = listexpect.splitlines()
        lines_actual = listactual.splitlines()
        for line in lines_expect:
            self.assertTrue(line in lines_actual, 'Missing entry %s in output %s' % (line, listactual))
        for line in lines_actual:
            self.assertTrue(line in lines_expect, 'Superflous entry %s in output %s' % (line, listactual))

    def assertEqualDiffs(self, expected, actual):
        "True if actual is similar enough to expected, minus svn properties"

        def filter_block(block):
            """removes property information that varies between systems, not relevant fo runit test"""
            newblock = []
            for line in block.splitlines():
                if re.search("[=+-\\@ ].*", line) == None:
                    break
                else:
                    # new svn versions use different labels for added
                    # files (working copy) vs (revision x)
                    fixedline = re.sub('\(revision [0-9]+\)', '(working copy)', line)
                    # svn 1.9 added (nonexistent)
                    fixedline = re.sub('\(nonexistent\)', '(working copy)', fixedline)
                    newblock.append(fixedline)
            return "\n".join(newblock)

        filtered_actual_blocks = []
        # A block starts with \nIndex, and the actual diff goes up to the first line starting with [a-zA-Z], e.g. "Properties changed:"
        for block in actual.split("\nIndex: "):
            if filtered_actual_blocks != []:
                # restore "Index: " removed by split()
                block = "Index: " + block
            block = filter_block(block)
            filtered_actual_blocks.append(block)
        expected_blocks = []
        for block in expected.split("\nIndex: "):
            if expected_blocks != []:
                block = "Index: " + block
            block = filter_block(block)
            expected_blocks.append(block)
        filtered = "\n".join(filtered_actual_blocks)
        self.assertEquals(set(expected_blocks), set(filtered_actual_blocks))

    def test_diff(self):
        client = SvnClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())

        self.assertEqualDiffs('Index: added.txt\n===================================================================\n--- added.txt\t(revision 0)\n+++ added.txt\t(revision 0)\n@@ -0,0 +1 @@\n+0123456789abcdef\n\\ No newline at end of file\nIndex: modified-fs.txt\n===================================================================\n--- modified-fs.txt\t(revision 3)\n+++ modified-fs.txt\t(working copy)\n@@ -0,0 +1 @@\n+0123456789abcdef\n\\ No newline at end of file\nIndex: modified.txt\n===================================================================\n--- modified.txt\t(revision 3)\n+++ modified.txt\t(working copy)\n@@ -0,0 +1 @@\n+0123456789abcdef\n\\ No newline at end of file',
                          client.get_diff().rstrip())

    def test_diff_relpath(self):
        client = SvnClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())

        self.assertEqualDiffs('Index: local/added.txt\n===================================================================\n--- local/added.txt\t(revision 0)\n+++ local/added.txt\t(revision 0)\n@@ -0,0 +1 @@\n+0123456789abcdef\n\\ No newline at end of file\nIndex: local/modified-fs.txt\n===================================================================\n--- local/modified-fs.txt\t(revision 3)\n+++ local/modified-fs.txt\t(working copy)\n@@ -0,0 +1 @@\n+0123456789abcdef\n\\ No newline at end of file\nIndex: local/modified.txt\n===================================================================\n--- local/modified.txt\t(revision 3)\n+++ local/modified.txt\t(working copy)\n@@ -0,0 +1 @@\n+0123456789abcdef\n\\ No newline at end of file', client.get_diff(basepath=os.path.dirname(self.local_path)).rstrip())

    def test_status(self):
        client = SvnClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertStatusListEqual('A       added.txt\nD       deleted.txt\nM       modified-fs.txt\n!       deleted-fs.txt\nM       modified.txt\n', client.get_status())

    def test_status_relpath(self):
        client = SvnClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertStatusListEqual('A       local/added.txt\nD       local/deleted.txt\nM       local/modified-fs.txt\n!       local/deleted-fs.txt\nM       local/modified.txt\n', client.get_status(basepath=os.path.dirname(self.local_path)))

    def test_status_untracked(self):
        client = SvnClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertStatusListEqual('?       added-fs.txt\nA       added.txt\nD       deleted.txt\nM       modified-fs.txt\n!       deleted-fs.txt\nM       modified.txt\n', client.get_status(untracked=True))


class SvnExportRepositoryClientTest(SvnClientTestSetups):

    @classmethod
    def setUpClass(self):
        SvnClientTestSetups.setUpClass()
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)

        self.basepath_export = os.path.join(self.root_directory, 'export')

    def tearDown(self):
        pass

    def test_export_repository(self):
        client = SvnClient(self.local_path)
        self.assertTrue(
          client.export_repository('',
            self.basepath_export)
        )

        self.assertTrue(os.path.exists(self.basepath_export + '.tar.gz'))
        self.assertFalse(os.path.exists(self.basepath_export + '.tar'))
        self.assertFalse(os.path.exists(self.basepath_export))


class SvnGetBranchesClientTest(SvnClientTestSetups):

    @classmethod
    def setUpClass(self):
        SvnClientTestSetups.setUpClass()
        client = SvnClient(self.local_path)
        client.checkout(self.local_url)

    # def tearDown(self):
    #     pass

    def test_get_remote_contents(self):
        self.assertEqual(['branches', 'tags', 'trunk'], get_remote_contents(self.local_root_url))

    def test_get_branches_non_canonical(self):
        remote_path = os.path.join(self.root_directory, "remote_nc")
        init_path = os.path.join(self.root_directory, "init_nc")
        local_path = os.path.join(self.root_directory, "local_nc")
        # create a "remote" repo
        subprocess.check_call("svnadmin create %s" % remote_path, shell=True, cwd=self.root_directory)
        local_root_url = "file://localhost/" + remote_path
        local_url = local_root_url + "/footest"
        # create an "init" repo to populate remote repo
        subprocess.check_call("svn checkout %s %s" % (local_root_url, init_path), shell=True, cwd=self.root_directory)
        for cmd in [
            "mkdir footest",
            "mkdir footest/foosub",
            "touch footest/foosub/fixed.txt",
            "svn add footest",
            "svn commit -m initial"]:
            subprocess.check_call(cmd, shell=True, cwd=init_path)
        client = SvnClient(local_path)
        client.checkout(local_url)
        self.assertEqual([], client.get_branches())

    def test_get_branches(self):
        client = SvnClient(self.local_path)

        self.assertEqual(['foo'], client.get_branches())

        # slyly create some empty branches
        subprocess.check_call("mkdir -p branches/foo2", shell=True, cwd=self.init_path)
        subprocess.check_call("mkdir -p branches/bar", shell=True, cwd=self.init_path)
        subprocess.check_call("svn add branches/foo2", shell=True, cwd=self.init_path)
        subprocess.check_call("svn add branches/bar", shell=True, cwd=self.init_path)
        subprocess.check_call("svn commit -m newbranches", shell=True, cwd=self.init_path)
        self.assertEqual([], client.get_branches(local_only=True))
        self.assertEqual(['bar', 'foo', 'foo2'], client.get_branches())

        # checkout branch foo
        local_path2 = os.path.join(self.root_directory, "local_foo")
        client = SvnClient(local_path2)
        client.checkout(self.local_root_url + '/branches/foo')
        self.assertEqual(['foo'], client.get_branches(local_only=True))



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

"""
bzr vcs support.
"""


from __future__ import absolute_import, print_function, unicode_literals
import os

import re
import email.utils  # For email parsing
import dateutil.parser  # Date string parsing

# first try python3, then python2
try:
    from urllib.request import url2pathname
except ImportError:
    from urllib2 import url2pathname

from vcstools.vcs_base import VcsClientBase, VcsError
from vcstools.common import sanitized, normalized_rel_path, \
    run_shell_command, ensure_dir_notexists


def _get_bzr_version():
    """Looks up bzr version by calling bzr --version.
    :raises: VcsError if bzr is not installed"""
    try:
        value, output, _ = run_shell_command('bzr --version',
                                             shell=True,
                                             us_env=True)
        if value == 0 and output is not None and len(output.splitlines()) > 0:
            version = output.splitlines()[0]
        else:
            raise VcsError("bzr --version returned %s," +
                           " maybe bzr is not installed" %
                           value)
    except VcsError as e:
        raise VcsError("Coud not determine whether bzr is installed: %s" % e)
    return version


class BzrClient(VcsClientBase):
    def __init__(self, path):
        """
        :raises: VcsError if bzr not detected
        """
        VcsClientBase.__init__(self, 'bzr', path)
        _get_bzr_version()

    @staticmethod
    def get_environment_metadata():
        metadict = {}
        try:
            metadict["version"] = _get_bzr_version()
        except:
            metadict["version"] = "no bzr installed"
        return metadict

    def get_url(self):
        """
        :returns: BZR URL of the branch (output of bzr info command),
        or None if it cannot be determined
        """
        result = None
        if self.detect_presence():
            cmd = 'bzr info %s' % self._path
            _, output, _ = run_shell_command(cmd, shell=True, us_env=True)
            matches = [l for l in output.splitlines() if l.startswith('  parent branch: ')]
            if matches:
                ppath = url2pathname(matches[0][len('  parent branch: '):])
                # when it can, bzr substitues absolute path for relative paths
                if (ppath is not None and os.path.isdir(ppath) and not os.path.isabs(ppath)):
                    result = os.path.abspath(os.path.join(os.getcwd(), ppath))
                else:
                    result = ppath
        return result

    def url_matches(self, url, url_or_shortcut):
        if super(BzrClient, self).url_matches(url, url_or_shortcut):
            return True
        # if we got a shortcut (e.g. launchpad url), we compare using
        # bzr info and return that one if result matches.
        result = False
        if url_or_shortcut is not None:
            cmd = 'bzr info %s' % url_or_shortcut
            value, output, _ = run_shell_command(cmd, shell=True, us_env=True)
            if value == 0:
                for line in output.splitlines():
                    sline = line.strip()
                    for prefix in ['shared repository: ',
                                   'repository branch: ',
                                   'branch root: ']:
                        if sline.startswith(prefix):
                            if super(BzrClient, self).url_matches(url, sline[len(prefix):]):
                                result = True
                                break
        return result

    @staticmethod
    def static_detect_presence(path):
        return os.path.isdir(os.path.join(path, '.bzr'))

    def checkout(self, url, version=None, verbose=False,
                 shallow=False, timeout=None):
        if url is None or url.strip() == '':
            raise ValueError('Invalid empty url : "%s"' % url)
        # bzr 2.5.1 fails if empty directory exists
        if not ensure_dir_notexists(self.get_path()):
            self.logger.error("Can't remove %s" % self.get_path())
            return False
        cmd = 'bzr branch'
        if version:
            cmd += ' -r %s' % version
        cmd += ' %s %s' % (url, self._path)
        value, _, msg = run_shell_command(cmd,
                                          shell=True,
                                          show_stdout=verbose,
                                          verbose=verbose)
        if value != 0:
            if msg:
                self.logger.error('%s' % msg)
            return False
        return True

    def update(self, version='', verbose=False, timeout=None):
        if not self.detect_presence():
            return False
        value, _, _ = run_shell_command("bzr pull",
                                        cwd=self._path,
                                        shell=True,
                                        show_stdout=True,
                                        verbose=verbose)
        if value != 0:
            return False
        # Ignore verbose param, bzr is pretty verbose on update anyway
        if version is not None and version != '':
            cmd = "bzr update -r %s" % (version)
        else:
            cmd = "bzr update"
        value, _, _ = run_shell_command(cmd,
                                        cwd=self._path,
                                        shell=True,
                                        show_stdout=True,
                                        verbose=verbose)
        if value == 0:
            return True
        return False

    def get_version(self, spec=None):
        """
        :param spec: (optional) revisionspec of desired version.  May
          be any revisionspec as returned by 'bzr help revisionspec',
          e.g. a tagname or 'revno:<number>'
        :returns: the current revision number of the repository. Or if
          spec is provided, the number of a revision specified by some
          token.
        """
        if self.detect_presence():
            if spec is not None:
                command = ['bzr log -r %s .' % sanitized(spec)]
                _, output, _ = run_shell_command(command,
                                                 shell=True,
                                                 cwd=self._path,
                                                 us_env=True)
                if output is None or output.strip() == '' or output.startswith("bzr:"):
                    return None
                else:
                    matches = [l for l in output.split('\n') if l.startswith('revno: ')]
                    if len(matches) == 1:
                        return matches[0].split()[1]
            else:
                _, output, _ = run_shell_command('bzr revno --tree',
                                                 shell=True,
                                                 cwd=self._path,
                                                 us_env=True)
                return output.strip()

    def get_current_version_label(self):
        # url contains branch information
        return None

    def get_remote_version(self, fetch=False):
        # Not sure how to get any useful information from bzr about this,
        # since bzr has no globally unique IDs
        return None

    def get_diff(self, basepath=None):
        response = None
        if basepath is None:
            basepath = self._path
        if self.path_exists():
            rel_path = sanitized(normalized_rel_path(self._path, basepath))
            command = "bzr diff %s" % rel_path
            command += " -p1 --prefix %s/:%s/" % (rel_path, rel_path)
            _, response, _ = run_shell_command(command, shell=True, cwd=basepath)
        return response

    def get_affected_files(self, revision):
        cmd = "bzr status -c {0} -S -V".format(
            revision)

        code, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)

        affected = []
        if code == 0:
            for filename in output.splitlines():
                affected.append(filename.split(" ")[2])
        return affected

    def get_log(self, relpath=None, limit=None):
        response = []

        if relpath is None:
            relpath = ''

        # Compile regexes
        id_regex = re.compile('^revno: ([0-9]+)$', flags=re.MULTILINE)
        committer_regex = re.compile('^committer: (.+)$', flags=re.MULTILINE)
        timestamp_regex = re.compile('^timestamp: (.+)$', flags=re.MULTILINE)
        message_regex = re.compile('^  (.+)$', flags=re.MULTILINE)

        if self.path_exists() and os.path.exists(os.path.join(self._path, relpath)):
            # Get the log
            limit_cmd = (("--limit=%d" % (int(limit))) if limit else "")
            command = "bzr log %s %s" % (sanitized(relpath), limit_cmd)
            return_code, text_response, stderr = run_shell_command(command, shell=True, cwd=self._path)
            if return_code == 0:
                revno_match = id_regex.findall(text_response)
                committer_match = committer_regex.findall(text_response)
                timestamp_match = timestamp_regex.findall(text_response)
                message_match = message_regex.findall(text_response)

                # Extract the entries
                for revno, committer, timestamp, message in zip(revno_match,
                                                                committer_match,
                                                                timestamp_match,
                                                                message_match):
                    author, email_address = email.utils.parseaddr(committer)
                    date = dateutil.parser.parse(timestamp)
                    log_data = {'id': revno,
                                'author': author,
                                'email': email_address,
                                'message': message,
                                'date': date}

                    response.append(log_data)

        return response

    def get_status(self, basepath=None, untracked=False):
        response = None
        if basepath is None:
            basepath = self._path
        if self.path_exists():
            rel_path = normalized_rel_path(self._path, basepath)
            command = "bzr status %s -S" % sanitized(rel_path)
            if not untracked:
                command += " -V"
            _, response, _ = run_shell_command(command, shell=True, cwd=basepath)
            response_processed = ""
            for line in response.split('\n'):
                if len(line.strip()) > 0:
                    response_processed += line[0:4] + rel_path + '/'
                    response_processed += line[4:] + '\n'
            response = response_processed
        return response

    def get_branches(self, local_only=False):
        # see http://doc.bazaar.canonical.com/beta/en/user-guide/shared_repository_layouts.html
        # the 'bzr branches' command exists, but is not useful here (too many assumptions)
        # Else bazaar branches are equivalent to forks in git and hg
        # such branches (forks) on launchpad could be retrieved using
        # the launchpadlib, but the API is probably not stable.
        raise NotImplementedError("get_branches is not implemented for bzr")

    def export_repository(self, version, basepath):
        # execute the bzr export cmd
        cmd = 'bzr export --format=tgz {0} '.format(basepath + '.tar.gz')
        cmd += '{0}'.format(version)
        result, _, _ = run_shell_command(cmd, shell=True, cwd=self._path)
        if result:
            return False
        return True

BZRClient = BzrClient



from __future__ import absolute_import, print_function, unicode_literals
import unittest
from mock import Mock

import vcstools.vcs_abstraction
from vcstools.vcs_abstraction import register_vcs, get_registered_vcs_types, \
    get_vcs

from vcstools import get_vcs_client


class TestVcsAbstraction(unittest.TestCase):

    def test_register_vcs(self):
        try:
            backup = vcstools.vcs_abstraction._VCS_TYPES
            vcstools.vcs_abstraction._VCS_TYPES = {}
            self.assertEqual([], get_registered_vcs_types())
            mock_class = Mock()
            register_vcs('foo', mock_class)
            self.assertEqual(['foo'], get_registered_vcs_types())
        finally:
            vcstools.vcs_abstraction._VCS_TYPES = backup

    def test_get_vcs(self):
        try:
            backup = vcstools.vcs_abstraction._VCS_TYPES
            vcstools.vcs_abstraction._VCS_TYPES = {}
            self.assertEqual([], get_registered_vcs_types())
            mock_class = Mock()
            register_vcs('foo', mock_class)
            self.assertEqual(mock_class, get_vcs('foo'))
            self.assertRaises(ValueError, get_vcs, 'bar')
        finally:
            vcstools.vcs_abstraction._VCS_TYPES = backup

    def test_get_vcs_client(self):
        try:
            backup = vcstools.vcs_abstraction._VCS_TYPES
            vcstools.vcs_abstraction._VCS_TYPES = {}
            self.assertEqual([], get_registered_vcs_types())
            mock_class = Mock()
            mock_instance = Mock()
            # mock __init__ constructor
            mock_class.return_value = mock_instance
            register_vcs('foo', mock_class)
            self.assertEqual(mock_instance, get_vcs_client('foo', 'foopath'))
            self.assertRaises(ValueError, get_vcs_client, 'bar', 'barpath')
        finally:
            vcstools.vcs_abstraction._VCS_TYPES = backup



#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, print_function, unicode_literals

import os
import io
import unittest
import subprocess
import tempfile
import shutil

from vcstools.hg import HgClient


os.environ['EMAIL'] = 'Your Name <name@example.com>'


class HGClientTestSetups(unittest.TestCase):

    @classmethod
    def setUpClass(self):

        self.root_directory = tempfile.mkdtemp()
        self.directories = dict(setUp=self.root_directory)
        self.remote_path = os.path.join(self.root_directory, "remote")
        os.makedirs(self.remote_path)

        # create a "remote" repo
        for cmd in [
                "hg init",
                "touch fixed.txt",
                "hg add fixed.txt",
                "hg commit -m initial"]:
            subprocess.check_call(cmd, shell=True, cwd=self.remote_path)

        po = subprocess.Popen("hg log --template '{node|short}' -l1", shell=True, cwd=self.remote_path, stdout=subprocess.PIPE)
        self.local_version_init = po.stdout.read().decode('UTF-8').rstrip("'").lstrip("'")
        # in hg, tagging creates an own changeset, so we need to fetch version before tagging
        subprocess.check_call("hg tag test_tag", shell=True, cwd=self.remote_path)

        # files to be modified in "local" repo
        for cmd in [
                "touch modified.txt",
                "touch modified-fs.txt",
                "hg add modified.txt modified-fs.txt",
                "hg commit -m initial"]:
            subprocess.check_call(cmd, shell=True, cwd=self.remote_path)

        po = subprocess.Popen("hg log --template '{node|short}' -l1", shell=True, cwd=self.remote_path, stdout=subprocess.PIPE)
        self.local_version_second = po.stdout.read().decode('UTF-8').rstrip("'").lstrip("'")

        for cmd in [
                "touch deleted.txt",
                "touch deleted-fs.txt",
                "hg add deleted.txt deleted-fs.txt",
                "hg commit -m modified"]:
            subprocess.check_call(cmd, shell=True, cwd=self.remote_path)

        po = subprocess.Popen("hg log --template '{node|short}' -l1", shell=True, cwd=self.remote_path, stdout=subprocess.PIPE)
        self.local_version = po.stdout.read().decode('UTF-8').rstrip("'").lstrip("'")

        self.local_path = os.path.join(self.root_directory, "local")
        self.local_url = self.remote_path

        # create a hg branch
        for cmd in [
                "hg branch test_branch",
                "touch test.txt",
                "hg add test.txt",
                "hg commit -m test"]:
            subprocess.check_call(cmd, shell=True, cwd=self.remote_path)

        po = subprocess.Popen("hg log --template '{node|short}' -l1", shell=True, cwd=self.remote_path, stdout=subprocess.PIPE)
        self.branch_version = po.stdout.read().decode('UTF-8').rstrip("'").lstrip("'")


    @classmethod
    def tearDownClass(self):
        for d in self.directories:
            shutil.rmtree(self.directories[d])

    def tearDown(self):
        if os.path.exists(self.local_path):
            shutil.rmtree(self.local_path)


class HGClientTest(HGClientTestSetups):

    def test_get_url_by_reading(self):
        url = self.local_url
        client = HgClient(self.local_path)
        client.checkout(url, self.local_version)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_url(), self.local_url)
        self.assertEqual(client.get_version(), self.local_version)
        self.assertEqual(client.get_version(self.local_version_init[0:6]), self.local_version_init)
        self.assertEqual(client.get_version("test_tag"), self.local_version_init)

    def test_get_url_nonexistant(self):
        local_path = "/tmp/dummy"
        client = HgClient(local_path)
        self.assertEqual(client.get_url(), None)

    def test_get_type_name(self):
        local_path = "/tmp/dummy"
        client = HgClient(local_path)
        self.assertEqual(client.get_vcs_type_name(), 'hg')

    def test_checkout(self):
        url = self.local_url
        client = HgClient(self.local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), self.local_path)
        self.assertEqual(client.get_url(), url)
        self.assertEqual(client.get_version(), self.local_version)

    def test_checkout_dir_exists(self):
        url = self.remote_path
        client = HgClient(self.local_path)
        self.assertFalse(client.path_exists())
        os.makedirs(self.local_path)
        self.assertTrue(client.checkout(url))
        # non-empty
        self.assertFalse(client.checkout(url))

    def test_checkout_emptystringversion(self):
        # special test to check that version '' means the same as None
        url = self.local_url
        client = HgClient(self.local_path)
        self.assertTrue(client.checkout(url, ''))
        self.assertEqual(client.get_version(), self.local_version)

    # test for #3497
    def test_checkout_into_subdir_without_existing_parent(self):
        local_path = os.path.join(self.local_path, "nonexistant_subdir")
        url = self.local_url
        client = HgClient(local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), local_path)
        self.assertEqual(client.get_url(), url)

    def test_checkout_specific_version_and_update(self):
        url = self.local_url
        version = self.local_version
        client = HgClient(self.local_path)
        self.assertFalse(client.path_exists())
        self.assertFalse(client.detect_presence())
        self.assertFalse(client.detect_presence())
        self.assertTrue(client.checkout(url, version))
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEqual(client.get_path(), self.local_path)
        self.assertEqual(client.get_url(), url)
        self.assertEqual(client.get_version(), version)

        new_version = self.local_version_second
        self.assertTrue(client.update(new_version))
        self.assertEqual(client.get_version(), new_version)

        self.assertTrue(client.update())
        self.assertEqual(client.get_version(), self.local_version)

        self.assertTrue(client.update(new_version))
        self.assertEqual(client.get_version(), new_version)

        self.assertTrue(client.update(''))
        self.assertEqual(client.get_version(), self.local_version)

    def test_get_current_version_label(self):
        url = self.local_url
        version = self.local_version
        client = HgClient(self.local_path)
        client.checkout(url, version='test_tag')
        self.assertEqual(client.get_current_version_label(), 'default')
        client.update(version='default')
        self.assertEqual(client.get_current_version_label(), 'default')
        client.update(version='test_branch')
        self.assertEqual(client.get_current_version_label(), 'test_branch')


    def test_get_remote_version(self):
        url = self.local_url
        version = self.local_version
        client = HgClient(self.local_path)
        client.checkout(url)
        self.assertEqual(client.get_remote_version(fetch=True), self.local_version)
        client.checkout(url, version='test_tag')
        self.assertEqual(client.get_remote_version(fetch=True), self.local_version)
        client.update(version='default')
        self.assertEqual(client.get_remote_version(fetch=True), self.local_version)
        client.update(version='test_branch')
        self.assertEqual(client.get_remote_version(fetch=True), self.branch_version)

    def testDiffClean(self):
        client = HgClient(self.remote_path)
        self.assertEquals('', client.get_diff())

    def testStatusClean(self):
        client = HgClient(self.remote_path)
        self.assertEquals('', client.get_status())

    def test_get_environment_metadata(self):
        # Verify that metadata is generated
        directory = tempfile.mkdtemp()
        self.directories['local'] = directory
        local_path = os.path.join(directory, "local")
        client = HgClient(local_path)
        self.assertTrue('version' in client.get_environment_metadata())


class HGClientLogTest(HGClientTestSetups):

    @classmethod
    def setUpClass(self):
        HGClientTestSetups.setUpClass()
        client = HgClient(self.local_path)
        client.checkout(self.local_url)

    def test_get_log_defaults(self):
        client = HgClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log()
        self.assertEquals(4, len(log))
        self.assertEquals('modified', log[0]['message'])
        for key in ['id', 'author', 'email', 'date', 'message']:
            self.assertTrue(log[0][key] is not None, key)

    def test_get_log_limit(self):
        client = HgClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log(limit=1)
        self.assertEquals(1, len(log))
        self.assertEquals('modified', log[0]['message'])

    def test_get_log_path(self):
        client = HgClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log(relpath='fixed.txt')
        self.assertEquals('initial', log[0]['message'])


class HGAffectedFilesTest(HGClientTestSetups):

    @classmethod
    def setUpClass(self):
        HGClientTestSetups.setUpClass()
        client = HgClient(self.local_path)
        client.checkout(self.local_url)

    def test_get_log_defaults(self):
        client = HgClient(self.local_path)
        client.checkout(self.local_url)
        log = client.get_log(limit=1)[0]
        affected = client.get_affected_files(log['id'])

        self.assertEqual(sorted(['deleted-fs.txt', 'deleted.txt']),
                         sorted(affected))



class HGDiffStatClientTest(HGClientTestSetups):

    @classmethod
    def setUpClass(self):
        HGClientTestSetups.setUpClass()
        url = self.local_url
        client = HgClient(self.local_path)
        client.checkout(url)
        # after setting up "local" repo, change files and make some changes
        subprocess.check_call("rm deleted-fs.txt", shell=True, cwd=self.local_path)
        subprocess.check_call("hg rm deleted.txt", shell=True, cwd=self.local_path)
        f = io.open(os.path.join(self.local_path, "modified.txt"), 'a')
        f.write('0123456789abcdef')
        f.close()
        f = io.open(os.path.join(self.local_path, "modified-fs.txt"), 'a')
        f.write('0123456789abcdef')
        f.close()
        f = io.open(os.path.join(self.local_path, "added-fs.txt"), 'w')
        f.write('0123456789abcdef')
        f.close()
        f = io.open(os.path.join(self.local_path, "added.txt"), 'w')
        f.write('0123456789abcdef')
        f.close()
        subprocess.check_call("hg add added.txt", shell=True, cwd=self.local_path)

    def tearDown(self):
        pass

    def test_diff(self):

        client = HgClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEquals('diff --git ./added.txt ./added.txt\nnew file mode 100644\n--- /dev/null\n+++ ./added.txt\n@@ -0,0 +1,1 @@\n+0123456789abcdef\n\\ No newline at end of file\ndiff --git ./deleted.txt ./deleted.txt\ndeleted file mode 100644\ndiff --git ./modified-fs.txt ./modified-fs.txt\n--- ./modified-fs.txt\n+++ ./modified-fs.txt\n@@ -0,0 +1,1 @@\n+0123456789abcdef\n\\ No newline at end of file\ndiff --git ./modified.txt ./modified.txt\n--- ./modified.txt\n+++ ./modified.txt\n@@ -0,0 +1,1 @@\n+0123456789abcdef\n\\ No newline at end of file', client.get_diff())

    def test_diff_relpath(self):

        client = HgClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())

        self.assertEquals('diff --git local/added.txt local/added.txt\nnew file mode 100644\n--- /dev/null\n+++ local/added.txt\n@@ -0,0 +1,1 @@\n+0123456789abcdef\n\\ No newline at end of file\ndiff --git local/deleted.txt local/deleted.txt\ndeleted file mode 100644\ndiff --git local/modified-fs.txt local/modified-fs.txt\n--- local/modified-fs.txt\n+++ local/modified-fs.txt\n@@ -0,0 +1,1 @@\n+0123456789abcdef\n\\ No newline at end of file\ndiff --git local/modified.txt local/modified.txt\n--- local/modified.txt\n+++ local/modified.txt\n@@ -0,0 +1,1 @@\n+0123456789abcdef\n\\ No newline at end of file', client.get_diff(basepath=os.path.dirname(self.local_path)))

    def test_get_version_modified(self):
        client = HgClient(self.local_path)
        self.assertFalse(client.get_version().endswith('+'))

    def test_status(self):
        client = HgClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEquals('M modified-fs.txt\nM modified.txt\nA added.txt\nR deleted.txt\n! deleted-fs.txt\n', client.get_status())

    def test_status_relpath(self):
        client = HgClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEquals('M local/modified-fs.txt\nM local/modified.txt\nA local/added.txt\nR local/deleted.txt\n! local/deleted-fs.txt\n', client.get_status(basepath=os.path.dirname(self.local_path)))

    def testStatusUntracked(self):
        client = HgClient(self.local_path)
        self.assertTrue(client.path_exists())
        self.assertTrue(client.detect_presence())
        self.assertEquals('M modified-fs.txt\nM modified.txt\nA added.txt\nR deleted.txt\n! deleted-fs.txt\n? added-fs.txt\n', client.get_status(untracked=True))

    def test_hg_diff_path_change_None(self):
        from vcstools.hg import _hg_diff_path_change
        self.assertEqual(_hg_diff_path_change(None, '/tmp/dummy'), None)


class HGRemoteFetchTest(HGClientTestSetups):

    def test_get_remote_version(self):
        url = self.local_url
        version = self.local_version
        client = HgClient(self.local_path)
        client.checkout(url, version='default')
        self.assertEqual(client.get_remote_version(fetch=True), self.local_version)
        self.assertEqual(client.get_version(), self.local_version)

        for cmd in [
                "hg checkout default",
                "touch remote_new.txt",
                "hg add remote_new.txt",
                "hg commit -m remote_new"]:
            subprocess.check_call(cmd, shell=True, cwd=self.remote_path)
        po = subprocess.Popen("hg log --template '{node|short}' -l1", shell=True, cwd=self.remote_path, stdout=subprocess.PIPE)
        remote_new_version = po.stdout.read().decode('UTF-8').rstrip("'").lstrip("'")
        self.assertNotEqual(self.local_version, remote_new_version)

        self.assertEqual(client.get_remote_version(fetch=False), self.local_version)
        self.assertEqual(client.get_remote_version(fetch=True), remote_new_version)
        self.assertEqual(client.get_remote_version(fetch=False), remote_new_version)
        self.assertEqual(client.get_version(), self.local_version)


class HGExportRepositoryClientTest(HGClientTestSetups):

    @classmethod
    def setUpClass(self):
        HGClientTestSetups.setUpClass()
        url = self.local_url
        client = HgClient(self.local_path)
        client.checkout(url)

        self.basepath_export = os.path.join(self.root_directory, 'export')

    def tearDown(self):
        pass

    def test_export_repository(self):
        client = HgClient(self.local_path)
        self.assertTrue(
          client.export_repository(self.local_version, self.basepath_export)
        )

        self.assertTrue(os.path.exists(self.basepath_export + '.tar.gz'))
        self.assertFalse(os.path.exists(self.basepath_export + '.tar'))
        self.assertFalse(os.path.exists(self.basepath_export))


class HGGetBranchesClientTest(HGClientTestSetups):

    @classmethod
    def setUpClass(self):
        HGClientTestSetups.setUpClass()
        url = self.local_url
        client = HgClient(self.local_path)
        client.checkout(url)

    def tearDown(self):
        pass

    def test_get_branches(self):
        client = HgClient(self.local_path)
        # Make a local branch
        subprocess.check_call('hg branch test_branch2', shell=True,
                              cwd=self.local_path, stdout=subprocess.PIPE)
        subprocess.check_call('hg commit -m "Making test_branch2"', shell=True,
                              cwd=self.local_path, stdout=subprocess.PIPE)
        self.assertEqual(client.get_branches(), ['test_branch2', 'test_branch', 'default'])

        # Make a remote branch
        subprocess.check_call('hg branch remote_branch', shell=True,
                              cwd=self.remote_path, stdout=subprocess.PIPE)
        subprocess.check_call("touch fixed.txt", shell=True,
                              cwd=self.remote_path)
        subprocess.check_call("hg add fixed.txt", shell=True,
                              cwd=self.remote_path)
        subprocess.check_call("hg commit -m initial", shell=True,
                              cwd=self.remote_path)
        self.assertEqual(client.get_branches(local_only=True), ['test_branch2', 'test_branch', 'default'])
        self.assertEqual(client.get_branches(), ['remote_branch', 'test_branch2', 'test_branch', 'default'])



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
"""
hg vcs support.

using ui object to redirect output into a string
"""


from __future__ import absolute_import, print_function, unicode_literals
import os
import sys

import gzip

import dateutil.parser  # For parsing date strings

from vcstools.vcs_base import VcsClientBase, VcsError
from vcstools.common import sanitized, normalized_rel_path, run_shell_command


def _get_hg_version():
    """Looks up hg version by calling hg --version.
    :raises: VcsError if hg is not installed"""
    try:
        value, output, _ = run_shell_command('hg --version',
                                             shell=True,
                                             us_env=True)
        if value == 0 and output is not None and len(output.splitlines()) > 0:
            version = output.splitlines()[0]
        else:
            raise VcsError("hg --version returned %s, output '%s', maybe hg is not installed" % (value, output))
    except VcsError as e:
        raise VcsError("Could not determine whether hg is installed %s" % e)
    return version


# hg diff cannot seem to be persuaded to accept a different prefix for filenames
def _hg_diff_path_change(diff, path):
    """
    Parses hg diff result and changes the filename prefixes.
    """
    if diff is None:
        return None
    INIT = 0
    INDIFF = 1
    # small state machine makes sure we never touch anything inside
    # the actual diff
    state = INIT

    s_list = [line for line in diff.split(os.linesep)]
    lines = []
    for line in s_list:
        if line.startswith("diff"):
            state = INIT
        if state == INIT:
            if line.startswith("@@"):
                state = INDIFF
                newline = line
            else:
                if line.startswith("---") and not line.startswith("--- /dev/null"):
                    newline = "--- %s%s" % (path, line[5:])
                elif line.startswith("+++") and not line.startswith("+++ /dev/null"):
                    newline = "+++ %s%s" % (path, line[5:])
                elif line.startswith("diff --git"):
                    # first replacing b in case path starts with a/
                    newline = line.replace(" b/", " " + path + "/", 1)
                    newline = newline.replace(" a/", " " + path + "/", 1)
                else:
                    newline = line
        else:
            newline = line
        if newline != '':
            lines.append(newline)
    result = "\n".join(lines)
    return result


class HgClient(VcsClientBase):

    def __init__(self, path):
        """
        :raises: VcsError if hg not detected
        """
        VcsClientBase.__init__(self, 'hg', path)
        _get_hg_version()

    @staticmethod
    def get_environment_metadata():
        metadict = {}
        try:
            metadict["version"] = '%s' % _get_hg_version()
        except:
            metadict["version"] = "no mercurial installed"
        return metadict

    def get_url(self):
        """
        :returns: HG URL of the directory path. (output of hg paths
        command), or None if it cannot be determined
        """
        if self.detect_presence():
            cmd = "hg paths default"
            _, output, _ = run_shell_command(cmd,
                                             shell=True,
                                             cwd=self._path,
                                             us_env=True)
            return output.rstrip()
        return None

    @staticmethod
    def static_detect_presence(path):
        return os.path.isdir(os.path.join(path, '.hg'))

    def checkout(self, url, version='', verbose=False,
                 shallow=False, timeout=None):
        if url is None or url.strip() == '':
            raise ValueError('Invalid empty url : "%s"' % url)
        # make sure that the parent directory exists for #3497
        base_path = os.path.split(self.get_path())[0]
        try:
            os.makedirs(base_path)
        except OSError:
            # OSError thrown if directory already exists this is ok
            pass
        cmd = "hg clone %s %s" % (sanitized(url), self._path)
        value, _, msg = run_shell_command(cmd,
                                          shell=True,
                                          no_filter=True)
        if value != 0:
            if msg:
                sys.logger.error('%s' % msg)
            return False
        if version is not None and version.strip() != '':
            cmd = "hg checkout %s" % sanitized(version)
            value, _, msg = run_shell_command(cmd,
                                              cwd=self._path,
                                              shell=True,
                                              no_filter=True)
            if value != 0:
                if msg:
                    sys.stderr.write('%s\n' % msg)
                return False
        return True

    def update(self, version='', verbose=False, timeout=None):
        verboseflag = ''
        if verbose:
            verboseflag = '--verbose'
        if not self.detect_presence():
            sys.stderr.write("Error: cannot update non-existing directory\n")
            return True
        if not self._do_pull():
            return False
        if version is not None and version.strip() != '':
            cmd = "hg checkout %s %s" % (verboseflag, sanitized(version))
        else:
            cmd = "hg update %s --config ui.merge=internal:fail" % verboseflag
        value, _, _ = run_shell_command(cmd,
                                        cwd=self._path,
                                        shell=True,
                                        no_filter=True)
        if value != 0:
            return False
        return True

    def get_version(self, spec=None):
        """
        :param spec: (optional) token for identifying version. spec can be
          a whatever is allowed by 'hg log -r', e.g. a tagname, sha-ID,
          revision-number
        :returns: the current SHA-ID of the repository. Or if spec is
          provided, the SHA-ID of a revision specified by some
          token.
        """
        # detect presence only if we need path for cwd in popen
        if spec is not None:
            if self.detect_presence():
                command = 'hg log -r %s' % sanitized(spec)
                repeated = False
                output = ''
                # we repeat the call once after pullin if necessary
                while output == '':
                    _, output, _ = run_shell_command(command,
                                                     shell=True,
                                                     cwd=self._path,
                                                     us_env=True)
                    if (output.strip() != '' and
                            not output.startswith("abort") or
                            repeated is True):

                        matches = [l for l in output.splitlines() if l.startswith('changeset: ')]
                        if len(matches) == 1:
                            return matches[0].split(':')[2]
                        else:
                            sys.stderr.write("Warning: found several candidates for hg spec %s" % spec)
                        break
                    self._do_pull()
                    repeated = True
            return None
        else:
            command = 'hg identify -i %s' % self._path
            _, output, _ = run_shell_command(command, shell=True, us_env=True)
            if output is None or output.strip() == '' or output.startswith("abort"):
                return None
            # hg adds a '+' to the end if there are uncommited
            # changes, inconsistent to hg log
            return output.strip().rstrip('+')

    def get_current_version_label(self):
        """
        :param spec: (optional) spec can be what 'svn info --help'
          allows, meaning a revnumber, {date}, HEAD, BASE, PREV, or
          COMMITTED.
        :returns: current revision number of the repository. Or if spec
          provided, the number of a revision specified by some
          token.
        """
        return self.get_branch()

    def get_branch(self):
        if self.path_exists():
            command = "hg branch --repository %s" % self.get_path()
            _, output, _ = run_shell_command(command, shell=True)
            if output is not None:
                return output.strip()
        return None

    def get_remote_version(self, fetch=False):
        if fetch:
            self._do_pull(filter=True)
        # use local information only
        result = self.get_log(limit=1)
        if (len(result) == 1 and 'id' in result[0]):
            return result[0]['id']
        return None

    def get_diff(self, basepath=None):
        response = None
        if basepath is None:
            basepath = self._path
        if self.path_exists():
            rel_path = normalized_rel_path(self._path, basepath)
            command = "hg diff -g %(path)s --repository %(path)s" % {'path': sanitized(rel_path)}
            _, response, _ = run_shell_command(command, shell=True, cwd=basepath)
            response = _hg_diff_path_change(response, rel_path)
        return response

    def get_affected_files(self, revision):
        cmd = "hg log -r %s --template '{files}'" % revision
        code, output, _ = run_shell_command(cmd, shell=True, cwd=self._path)
        affected = []
        if code == 0:
            affected = output.split(" ")
        return affected

    def get_log(self, relpath=None, limit=None):
        response = []

        if relpath is None:
            relpath = ''

        if self.path_exists() and os.path.exists(os.path.join(self._path, relpath)):
            # Get the log
            limit_cmd = (("--limit %d" % (int(limit))) if limit else "")
            HG_COMMIT_FIELDS = ['id', 'author', 'email', 'date', 'message']
            HG_LOG_FORMAT = '\x1f'.join(['{node|short}', '{author|person}',
                                         '{autor|email}', '{date|isodate}',
                                         '{desc}']) + '\x1e'

            command = "hg log %s -b %s --template '%s' %s" % (sanitized(relpath),
                                                              self.get_branch(),
                                                              HG_LOG_FORMAT,
                                                              limit_cmd)
            return_code, response_str, stderr = run_shell_command(command, shell=True, cwd=self._path)

            if return_code == 0:
                # Parse response
                response = response_str.strip('\n\x1e').split("\x1e")
                response = [row.strip().split("\x1f") for row in response]
                response = [dict(zip(HG_COMMIT_FIELDS, row)) for row in response]
                # Parse dates
                for entry in response:
                    entry['date'] = dateutil.parser.parse(entry['date'])

        return response

    def get_status(self, basepath=None, untracked=False):
        response = None
        if basepath is None:
            basepath = self._path
        if self.path_exists():
            rel_path = normalized_rel_path(self._path, basepath)
            # protect against shell injection
            command = "hg status %(path)s --repository %(path)s" % {'path': sanitized(rel_path)}
            if not untracked:
                command += " -mard"
            _, response, _ = run_shell_command(command,
                                               shell=True,
                                               cwd=basepath)
            if response is not None:
                if response.startswith("abort"):
                    raise VcsError("Probable Bug; Could not call %s, cwd=%s" % (command, basepath))
                if len(response) > 0 and response[-1] != '\n':
                    response += '\n'
        return response

    def export_repository(self, version, basepath):
        # execute the hg archive cmd
        cmd = 'hg archive -t tar -r {0} {1}.tar'.format(version, basepath)
        result, _, _ = run_shell_command(cmd, shell=True, cwd=self._path)
        if result:
            return False
        try:
            # gzip the tar file
            with open(basepath + '.tar', 'rb') as tar_file:
                gzip_file = gzip.open(basepath + '.tar.gz', 'wb')
                try:
                    gzip_file.writelines(tar_file)
                finally:
                    gzip_file.close()
        finally:
            # clean up
            os.remove(basepath + '.tar')
        return True

    def get_branches(self, local_only=False):
        if not local_only:
            self._do_pull()
        cmd = 'hg branches'
        result, out, _ = run_shell_command(cmd, shell=True, cwd=self._path,
                                           show_stdout=False)
        if result:
            return []
        branches = []
        for line in out.splitlines():
            line = line.strip()
            line = line.split()
            branches.append(line[0])
        return branches

    def _do_pull(self, filter=False):
        value, _, _ = run_shell_command("hg pull",
                                        cwd=self._path,
                                        shell=True,
                                        no_filter=not filter)
        return value == 0

# backwards compat
HGClient = HgClient



# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
"""
tar vcs support.

The implementation uses the "version" argument to indicate a subfolder
within a tarfile.  Hence one can organize sources by creating one
tarfile with a folder inside for each version.
"""


from __future__ import absolute_import, print_function, unicode_literals
import os
import tempfile
import shutil
import tarfile
import sys
import yaml
from vcstools.vcs_base import VcsClientBase, VcsError
from vcstools.common import urlretrieve_netrc, ensure_dir_notexists


__pychecker__ = 'unusednames=spec'

_METADATA_FILENAME = ".tar"


class TarClient(VcsClientBase):

    def __init__(self, path):
        """
        @raise VcsError if tar not detected
        """
        VcsClientBase.__init__(self, 'tar', path)
        self.metadata_path = os.path.join(self._path, _METADATA_FILENAME)

    @staticmethod
    def get_environment_metadata():
        metadict = {}
        metadict["version"] = 'tarfile version: %s' % tarfile.version
        return metadict

    def get_url(self):
        """
        :returns: TAR URL of the directory path (output of tar info
        command), or None if it cannot be determined
        """
        if self.detect_presence():
            with open(self.metadata_path, 'r') as metadata_file:
                metadata = yaml.load(metadata_file.read())
                if 'url' in metadata:
                    return metadata['url']
        return None

    @staticmethod
    def static_detect_presence(path):
        return os.path.isfile(os.path.join(path, _METADATA_FILENAME))

    def checkout(self, url, version='', verbose=False,
                 shallow=False, timeout=None):
        """
        untars tar at url to self.path.
        If version was given, only the subdirectory 'version' of the
        tar will end up in self.path.  Also creates a file next to the
        checkout named *.tar which is a yaml file listing origin url
        and version arguments.
        """
        if not ensure_dir_notexists(self.get_path()):
            self.logger.error("Can't remove %s" % self.get_path())
            return False
        tempdir = None
        result = False
        try:
            tempdir = tempfile.mkdtemp()
            if os.path.isfile(url):
                filename = url
            else:
                (filename, _) = urlretrieve_netrc(url)
                # print "filename", filename
            temp_tarfile = tarfile.open(filename, 'r:*')
            members = None  # means all members in extractall
            if version == '' or version is None:
                subdir = tempdir
                self.logger.warn("No tar subdirectory chosen via the 'version' argument for url: %s" % url)
            else:
                # getmembers lists all files contained in tar with
                # relative path
                subdirs = []
                members = []
                for m in temp_tarfile.getmembers():
                    if m.name.startswith(version + '/'):
                        members.append(m)
                    if m.name.split('/')[0] not in subdirs:
                        subdirs.append(m.name.split('/')[0])
                if not members:
                    raise VcsError("%s is not a subdirectory with contents in members %s" % (version, subdirs))
                subdir = os.path.join(tempdir, version)
            temp_tarfile.extractall(path=tempdir, members=members)

            if not os.path.isdir(subdir):
                raise VcsError("%s is not a subdirectory\n" % subdir)

            try:
                # os.makedirs(os.path.dirname(self._path))
                shutil.move(subdir, self._path)
            except Exception as ex:
                raise VcsError("%s failed to move %s to %s" % (ex, subdir, self._path))
            metadata = yaml.dump({'url': url, 'version': version})
            with open(self.metadata_path, 'w') as mdat:
                mdat.write(metadata)
            result = True

        except Exception as exc:
            self.logger.error("Tarball download unpack failed: %s" % str(exc))
        finally:
            if tempdir is not None and os.path.exists(tempdir):
                shutil.rmtree(tempdir)
        return result

    def update(self, version='', verbose=False, timeout=None):
        """
        Does nothing except returning true if tar exists in same
        "version" as checked out with vcstools.
        """
        if not self.detect_presence():
            return False
        if version != self.get_version():
            sys.stderr.write("Tarball Client does not support updating with different version '%s' != '%s'\n"
                             % (version, self.get_version()))
            return False

        return True

    def get_version(self, spec=None):

        if self.detect_presence():
            with open(self.metadata_path, 'r') as metadata_file:
                metadata = yaml.load(metadata_file.read())
                if 'version' in metadata:
                    return metadata['version']
        return None

    def get_current_version_label(self):
        # exploded tar has no local version
        return None

    def get_remote_version(self, fetch=False):
        # exploded tar has no remote version (not a required feature)
        return None

    def get_diff(self, basepath=None):
        return ''

    def get_status(self, basepath=None, untracked=False):
        return ''

    def export_repository(self, version, basepath):
        raise VcsError('export repository not implemented for extracted tars')


# backwards compatibility
TARClient = TarClient
